from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

from jigu.core.auth.transaction import StdTx, TxInfo
from jigu.core.sdk import Timestamp
from jigu.query.txinfo import TxInfosQuery
from jigu.util.serdes import JiguBox, JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S


@dataclass
class BlockHeader(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        block_id=S.OBJ,
        header=S.OBJECT(
            chain_id=S.STRING,
            height=S.STRING_INTEGER,
            time=Timestamp.__schema__,
            num_txs=S.STRING_INTEGER,
            total_txs=S.STRING_INTEGER,
            last_commit_hash=S.STRING,
            data_hash=S.STRING,
            next_validators_hash=S.STRING,
            consensus_hash=S.STRING,
            app_hash=S.STRING,
            proposer_address=S.STRING,
            last_block_id=S.OBJ,
        ),
    )

    chain_id: str
    height: int
    time: Timestamp
    num_txs: int
    total_txs: int
    hashes: dict
    proposer: str
    block_id: JiguBox
    last_block_id: JiguBox

    @classmethod
    def from_data(cls, data: dict) -> BlockHeader:
        header = data["header"]
        hashes = JiguBox(
            {
                "last_commit": header["last_commit_hash"],
                "data": header["data_hash"],
                "validators": header["validators_hash"],
                "next_validators": header["next_validators_hash"],
                "consensus": header["consensus_hash"],
                "app": header["app_hash"],
                "last_results": header["last_results_hash"],
                "evidence": header["evidence_hash"],
            }
        )
        return cls(
            chain_id=header["chain_id"],
            height=int(header["height"]),
            time=Timestamp.from_data(header["time"]),
            num_txs=int(header["num_txs"]),
            total_txs=int(header["total_txs"]),
            hashes=hashes,
            proposer=header["proposer_address"],
            block_id=JiguBox(data["block_id"]),
            last_block_id=JiguBox(header["last_block_id"]),
        )


# TODO: serialization currently doesn't match how it gets deserialized -- should
# we try to make it consistent? normal block fetching behavior gets txinfo instead of
# stdtx for the txs column.


@dataclass
class Block(BlockHeader, JsonDeserializable):

    __schema__ = S.OBJECT(
        block_meta=BlockHeader.__schema__,
        block=S.OBJECT(
            data=S.OBJECT(txs=S.OPTIONAL(S.ARRAY(S.STRING))),
            evidence={},
            last_commit={},
        ),
    )

    evidence: JiguBox  # TODO: improve and document
    last_commit: JiguBox  # TODO: improve and document
    txs: List[
        Union[str, StdTx, TxInfo, TxInfosQuery]
    ]  # ordered last for convenience of pretty printing

    @classmethod
    def from_data(cls, data: dict) -> Block:
        header = BlockHeader.from_data(data["block_meta"])
        txs = (
            data["block"]["data"]["txs"] or []
        )  # these will be Amino-encoded tx strings
        evidence = JiguBox(data["block"].get("evidence"))
        last_commit = JiguBox(data["block"].get("last_commit"))
        return cls(
            txs=txs, evidence=evidence, last_commit=last_commit, **header.__dict__
        )



from typing import Dict, Optional, Union

import jigu
from jigu.client.lcd.api import project
from jigu.client.lcd.api.modules import (
    AuthApi,
    BankApi,
    DistributionApi,
    GovApi,
    MarketApi,
    OracleApi,
    SlashingApi,
    StakingApi,
    SupplyApi,
    TreasuryApi
)
from jigu.client.lcd.api.tendermint import TendermintApi
from jigu.client.lcd.api.tx import TxApi
from jigu.client.lcd.lcdclient import LcdClient
from jigu.client.object_query import (
    AccountQuery,
    DenomQuery,
    ProposalQuery,
    ValidatorQuery,
    Wallet
)
from jigu.client.websocket import WebSocketClient
from jigu.core import AccAddress, Coins, ValAddress
from jigu.error import DenomNotFound
from jigu.listener import BlockListener, TxListener
from jigu.util.serdes import JiguBox
from jigu.util.validation import validate_acc_address, validate_val_address


class Terra(object):
    def __init__(
        self,
        chain_id: str,
        lcd_url: str,
        ws_url: str = None,
        gas_prices: Coins = None,
        gas_adjustment: Union[float, str] = "1.4",  # sensible defaults
    ):
        gas_prices = gas_prices or Coins(uluna="0.015")  # sensible defaults

        self.gas_prices = gas_prices
        self.gas_adjustment = gas_adjustment

        # LCD APIs
        self.lcd = LcdClient(self, lcd_url)

        # LCD module APIs
        self._auth = AuthApi(self)
        self._bank = BankApi(self)
        self._supply = SupplyApi(self)
        self._distribution = DistributionApi(self)
        self._staking = StakingApi(self)
        self._slashing = SlashingApi(self)
        self._oracle = OracleApi(self)
        self._market = MarketApi(self)
        self._treasury = TreasuryApi(self)
        self._gov = GovApi(self)

        # LCD lower-level APIs
        self._tendermint = TendermintApi(self)
        self._tx = TxApi(self)

        # LCD query APIs
        self._blocks = jigu.client.object_query.BlocksQuery(self)

        # if no chain_id, trust the node's chain_id
        if chain_id is None:
            # TODO: add warning if not same!
            self.chain_id = self.node_info()["node_info"]["network"]
        else:
            self.chain_id = chain_id

        # WebSocket APIs
        self.ws = WebSocketClient(self, ws_url)

    def __repr__(self):
        s = self
        return f"Terra('{s.chain_id}', '{s.lcd.url}')"

    def __str__(self):
        return f"{self.chain_id} via {self.lcd.url}"

    # Terra Core Module APIs provided through LCD:

    @property
    def auth(self):
        return self._auth

    @property
    def bank(self):
        return self._bank

    def supply(self, denom: Optional[str] = None):
        total_supply = self._supply.total()
        if denom is None:
            return total_supply
        if denom not in total_supply:
            raise DenomNotFound(
                f"denom '{denom}' was not found, avaialble denoms are: {total_supply.denoms}"
            )
        return project(total_supply, total_supply[denom])

    @property
    def distribution(self):
        return self._distribution

    @property
    def staking(self):
        return self._staking

    @property
    def slashing(self):
        return self._slashing

    @property
    def oracle(self):
        return self._oracle

    @property
    def market(self):
        return self._market

    @property
    def treasury(self):
        return self._treasury

    @property
    def gov(self):
        return self._gov

    @property
    def tx(self):
        return self._tx

    def is_connected(self) -> bool:
        """Checks that a connection can be made to the node specified, and has the same `chain_id`."""
        try:
            node_info = self.node_info()
            return self.chain_id == node_info["node_info"]["network"]
        except:  # not recommended, but if we run into any error, we are not connected.
            return False

    # lower-level APIs

    def is_syncing(self) -> bool:
        """Checks whether the node is currently syncing with the blockchain."""
        return self._tendermint.syncing()

    def node_info(self) -> Dict[str, dict]:
        """Get information about the node."""
        return self._tendermint.node_info()

    # Convenient Aliases for TX

    def tx_info(self, *args, **kwargs):
        return self._tx.tx_info(*args, **kwargs)

    def estimate_fee(self, *args, **kwargs):
        return self._tx.estimate_fee(*args, **kwargs)

    def broadcast(self, *args, **kwargs):
        return self._tx.broadcast(*args, **kwargs)

    # Object query based APIs

    def account(self, arg: AccAddress) -> AccountQuery:
        arg = validate_acc_address(arg)
        return AccountQuery(self, arg)

    def validator(self, arg: ValAddress) -> ValidatorQuery:
        arg = validate_val_address(arg)
        return ValidatorQuery(self, arg)

    def denom(self, arg: str) -> DenomQuery:
        return DenomQuery(self, arg)

    def wallet(self, arg: jigu.key.Key) -> Wallet:
        return Wallet(self, arg)

    def proposal(self, proposal_id: int) -> ProposalQuery:
        return ProposalQuery(self, proposal_id)

    ##

    @property
    def blocks(self):
        """Blocks querying API."""
        return self._blocks

    def block(self, height=None):
        """Get the block at a specific height."""
        return self._blocks.at(height)

    def delegates(self, height=None):
        all_validators = self.staking.validators()
        vset = self._tendermint.validator_set(height=height)
        # join validators using pubkey
        by_pubkey = {x["pub_key"]: x for x in vset}
        return JiguBox(
            {
                v.operator_address: {"info": v, **by_pubkey[v.consensus_pubkey]}
                for v in all_validators
                if v.consensus_pubkey in by_pubkey
            }
        )

    ## WebSockets

    def tx_listener(self, query: dict = None):
        def decorator(f):
            return TxListener(self, func=f, query=query)

        return decorator

    def block_listener(self, query: dict = None):
        def decorator(f):
            return BlockListener(self, func=f, query=query)

        return decorator



from __future__ import annotations

from jigu.core import AccAddress, Coins
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address

__all__ = ["Input", "Output"]


class MultiSendIO(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(address=S.ACC_ADDRESS, coins=Coins.__schema__)

    def __init__(self, address: AccAddress, coins: Coins):
        address = validate_acc_address(address)
        self.address = address
        self.coins = Coins(coins)

    @classmethod
    def from_data(cls, data: dict) -> MultiSendIO:
        return cls(address=data["address"], coins=Coins.from_data(data["coins"]))


class Input(MultiSendIO):
    def __repr__(self) -> str:
        return f"Input(address='{self.address}', coins={self.coins!r})"


class Output(MultiSendIO):
    def __repr__(self) -> str:
        return f"Output(address='{self.address}', coins={self.coins!r})"



import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.distribution import (
    MsgModifyWithdrawAddress,
    MsgWithdrawDelegationReward,
    MsgWithdrawValidatorCommission,
)
from jigu.util.validation import (
    InvalidAccAddress,
    InvalidValAddress,
    is_acc_address,
    is_val_address,
)
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for distribution messages."""
    return load_msg_examples(
        tdd,
        msg_types=[
            MsgModifyWithdrawAddress,
            MsgWithdrawDelegationReward,
            MsgWithdrawValidatorCommission,
        ],
    )


class TestMsgModifyWithdrawAddress:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgModifyWithdrawAddress]:
            assert_serdes_consistent(MsgModifyWithdrawAddress, m)
            assert_serdes_exact(MsgModifyWithdrawAddress, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgModifyWithdrawAddress.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgModifyWithdrawAddress, m)

    def test_matches_meta(self):
        assert MsgModifyWithdrawAddress.type == "distribution/MsgModifyWithdrawAddress"
        assert MsgModifyWithdrawAddress.action == "set_withdraw_address"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgModifyWithdrawAddress should validate `delegator_address` and `withdraw_address` to
        be correct AccAddress.
        """

        assume(not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgModifyWithdrawAddress(
                delegator_address=other, withdraw_address=acc_address
            )
        with pytest.raises(InvalidAccAddress):
            MsgModifyWithdrawAddress(
                delegator_address=acc_address, withdraw_address=other
            )
        A = MsgModifyWithdrawAddress(
            delegator_address=acc_address, withdraw_address=acc_address
        )
        assert A.delegator_address == acc_address


class TestMsgWithdrawDelegationReward:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgWithdrawDelegationReward]:
            assert_serdes_consistent(MsgWithdrawDelegationReward, m)
            assert_serdes_exact(MsgWithdrawDelegationReward, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgWithdrawDelegationReward.schema))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgWithdrawDelegationReward, m)

    def test_matches_meta(self):
        assert (
            MsgWithdrawDelegationReward.type
            == "distribution/MsgWithdrawDelegationReward"
        )
        assert MsgWithdrawDelegationReward.action == "withdraw_delegation_reward"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, val_address, other):
        """MsgWithdrawDelegationReward should validate `delegator_address` to be AccAddress
        and `withdraw_address` to be ValAddress.
        """

        assume(not is_acc_address(other) and not is_val_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgWithdrawDelegationReward(
                delegator_address=other, validator_address=val_address
            )
        with pytest.raises(InvalidValAddress):
            MsgWithdrawDelegationReward(
                delegator_address=acc_address, validator_address=other
            )
        A = MsgWithdrawDelegationReward(
            delegator_address=acc_address, validator_address=val_address
        )
        assert A.delegator_address == acc_address
        assert A.validator_address == val_address


class TestMsgWithdrawValidatorCommission:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgWithdrawValidatorCommission]:
            assert_serdes_consistent(MsgWithdrawValidatorCommission, m)
            assert_serdes_exact(MsgWithdrawValidatorCommission, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgWithdrawValidatorCommission.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgWithdrawValidatorCommission, m)

    def test_matches_meta(self):
        assert (
            MsgWithdrawValidatorCommission.type
            == "distribution/MsgWithdrawValidatorCommission"
        )
        assert MsgWithdrawValidatorCommission.action == "withdraw_validator_commission"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, val_address, other):
        """MsgWithdrawValidatorCommission should validate `validator_address` to be ValAddress.
        """

        assume(not is_val_address(other))
        with pytest.raises(InvalidValAddress):
            MsgWithdrawValidatorCommission(validator_address=other)
        A = MsgWithdrawValidatorCommission(validator_address=val_address)
        assert A.validator_address == val_address



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Union

from jigu.core import Coin, Dec, Timestamp, ValAddress
from jigu.core.denoms import uLuna
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = [
    "CommissionRates",
    "Commission",
    "DoNotModifyDesc",
    "Description",
    "Validator",
]


@dataclass
class CommissionRates(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        rate=Dec.__schema__, max_rate=Dec.__schema__, max_change_rate=Dec.__schema__
    )

    rate: Dec
    max_rate: Dec
    max_change_rate: Dec

    @classmethod
    def from_data(cls, data: dict) -> CommissionRates:
        return cls(
            rate=Dec(data["rate"]),
            max_rate=Dec(data["max_rate"]),
            max_change_rate=Dec(data["max_change_rate"]),
        )


@dataclass
class Commission(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        commission_rates=CommissionRates.__schema__, update_time=Timestamp.__schema__
    )

    rates: CommissionRates
    update_time: Timestamp

    def to_data(self) -> Dict[str, Union[CommissionRates, Timestamp]]:
        return {"commission_rates": self.rates, "update_time": self.update_time}

    @classmethod
    def from_data(cls, data: dict) -> Commission:
        return cls(
            rates=CommissionRates.from_data(data["commission_rates"]),
            update_time=Timestamp.from_data(data["update_time"]),
        )


DoNotModifyDesc = "[do-not-modify]"  # from cosmos


@dataclass
class Description(JsonSerializable, JsonDeserializable):
    __schema__ = S.OBJECT(
        moniker=S.STRING, identity=S.STRING, website=S.STRING, details=S.STRING
    )

    moniker: str = ""
    identity: str = ""
    website: str = ""
    details: str = ""

    @classmethod
    def do_not_modify(cls):
        return cls(DoNotModifyDesc, DoNotModifyDesc, DoNotModifyDesc, DoNotModifyDesc)

    @classmethod
    def from_data(cls, data) -> Description:
        return cls(data["moniker"], data["identity"], data["website"], data["details"])


@dataclass
class Validator(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        operator_address=S.VAL_ADDRESS,
        consensus_pubkey=S.STRING,
        tokens=S.STRING_INTEGER,
        jailed=S.BOOLEAN,
        status=S.INTEGER,  # this is converted to string
        delegator_shares=Dec.__schema__,
        description=Description.__schema__,
        unbonding_height=S.STRING_INTEGER,
        unbonding_time=Timestamp.__schema__,
        commission=Commission.__schema__,
        min_self_delegation=S.STRING_INTEGER,
    )

    operator_address: ValAddress
    consensus_pubkey: str
    jailed: bool
    status_code: int
    tokens: Coin
    delegator_shares: Coin
    description: Description
    unbonding_height: int
    unbonding_time: Timestamp
    commission: Commission
    min_self_delegation: int

    @property
    def status(self) -> str:
        """String version of `status_code`"""
        return ["unbonded", "unbonding", "bonded"][self.status_code]

    def to_data(self) -> dict:
        s = self
        d = dict(self.__dict__)
        del d["status_code"]
        d["status"] = s.status_code
        d["tokens"] = str(s.tokens.amount)
        d["delegator_shares"] = str(s.delegator_shares.amount)
        d["unbonding_height"] = str(s.unbonding_height)
        d["min_self_delegation"] = str(s.min_self_delegation)
        return d

    @classmethod
    def from_data(cls, data: dict) -> Validator:
        return cls(
            operator_address=data["operator_address"],
            consensus_pubkey=data["consensus_pubkey"],
            jailed=data["jailed"],
            status_code=data["status"],
            tokens=Coin(uLuna, data["tokens"]),
            delegator_shares=Coin(uLuna, data["delegator_shares"]),
            description=Description.from_data(data["description"]),
            unbonding_height=int(data["unbonding_height"]),
            unbonding_time=Timestamp.from_data(data["unbonding_time"]),
            commission=Commission.from_data(data["commission"]),
            min_self_delegation=int(data["min_self_delegation"]),
        )



from __future__ import annotations

import re
from dataclasses import dataclass
from decimal import Decimal
from typing import Callable, Dict, Generic, Iterable, List, TypeVar, Union

from jigu.error import DenomIncompatibleError
from jigu.util.serdes import JiguBox, JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

from .dec import Dec

T = TypeVar("T", int, Dec)  # Amount Type

__all__ = ["Coin", "Coins"]


@dataclass(frozen=True)
class Coin(JsonSerializable, JsonDeserializable, Generic[T]):

    __schema__ = S.OBJECT(
        denom=S.STRING, amount=S.ANY(S.STRING_INTEGER, Dec.__schema__)
    )

    denom: str
    amount: T  # all get converted to int or Dec

    def __post_init__(self):
        s = self
        if (
            isinstance(s.amount, Dec)
            or isinstance(s.amount, Decimal)
            or isinstance(s.amount, float)
            or (
                isinstance(s.amount, str)
                and "." in str(s.amount)
                or "E" in str(s.amount)
            )
        ):
            object.__setattr__(
                s, "amount", Dec(s.amount)
            )  # must do this due to immutability
        else:
            try:
                object.__setattr__(s, "amount", int(s.amount))
            except ValueError:
                raise ValueError(f"unacceptable value for Coin amount: {s.amount}")

    def __repr__(self) -> str:
        return f"Coin('{self.denom}', {self.amount!r})"

    @property
    def int_coin(self) -> Coin[int]:
        return Coin(self.denom, int(self.amount))

    @property
    def dec_coin(self) -> Coin[Dec]:
        return Coin(self.denom, Dec(self.amount))

    def __str__(self) -> str:
        return f"{self.amount}{self.denom}"

    @classmethod
    def from_str(cls, string: str) -> Coin:
        pattern = r"^(\-?[0-9]+(\.[0-9]+)?)([a-zA-Z]+)$"
        match = re.match(pattern, string)
        if match:
            return cls(match.group(3), match.group(1))
        else:
            raise ValueError(f"{string} could not be parsed into Coin")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Coin):
            return self.denom == other.denom and self.amount == other.amount
        else:
            return False

    def __add__(self, other) -> Union[Coin, Coins]:
        if other == 0:
            return Coin(self.denom, self.amount)
        if isinstance(other, Coins):
            return other + self
        if not isinstance(other, Coin):
            raise TypeError(
                f"unsupported operand types for +: 'Coin' and '{type(other)}'"
            )
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"unsupported Coin denoms for +: '{self.denom}' and '{other.denom}'"
            )
        cast = int
        if isinstance(self.amount, Dec) or isinstance(other.amount, Dec):
            cast = Dec
        return Coin(denom=self.denom, amount=cast(self.amount + other.amount))

    def __radd__(self, other) -> Union[Coin, Coins]:
        return self + other

    def __sub__(self, other: Coin) -> Coin[Union[int, Dec]]:
        if not isinstance(other, Coin):
            raise TypeError(
                f"unsupported operand types for -: 'Coin' and '{type(other)}'"
            )
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"unsupported Coin denoms for -: '{self.denom}' and '{other.denom}'"
            )
        cast = int
        if isinstance(self.amount, Dec) or isinstance(other.amount, Dec):
            cast = Dec
        return Coin(denom=self.denom, amount=cast(self.amount - other.amount))

    def __mul__(self, other) -> Coin[T]:
        return Coin(denom=self.denom, amount=(self.amount * other))

    def __rmul__(self, other) -> Coin[T]:
        return self * other

    def __truediv__(self, other) -> Coin[T]:
        cast = type(self.amount)
        return Coin(denom=self.denom, amount=cast(self.amount / other))

    def __floordiv__(self, other) -> Coin[T]:
        cast = type(self.amount)
        return Coin(denom=self.denom, amount=cast(self.amount // other))

    def __neg__(self) -> Coin[T]:
        return Coin(denom=self.denom, amount=(-self.amount))

    def __abs__(self) -> Coin[T]:
        return Coin(denom=self.denom, amount=abs(self.amount))

    def __pos__(self) -> Coin[T]:
        return abs(self)

    def __lt__(self, other: Coin) -> bool:
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"incompatible Coin denoms for <: '{self.denom}' and '{other.denom}'"
            )
        return self.amount < other.amount

    def __gt__(self, other: Coin) -> bool:
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"incompatible Coin denoms for >: '{self.denom}' and '{other.denom}'"
            )
        return self.amount > other.amount

    def __ge__(self, other: Coin) -> bool:
        return self > other or self == other

    def __le__(self, other: Coin) -> bool:
        return self < other or self == other

    def to_data(self) -> dict:
        return {"denom": str(self.denom), "amount": str(self.amount)}

    def _pretty_repr_(self, path="") -> str:
        return str(self)

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> Coin:
        return cls(denom=data["denom"], amount=data["amount"])


class Coins(JsonSerializable, JsonDeserializable, Generic[T]):

    __schema__ = S.ARRAY(Coin.__schema__)

    def __init__(self, coins: Iterable[Coin] = None, **denoms):
        if coins is None:
            coins = []
        coins = list(coins) + [Coin(d, a) for d, a in denoms.items()]
        self._cd = dict()
        for coin in list(coins):
            if self._cd.get(coin.denom, None) is None:
                self._cd[coin.denom] = Coin(coin.denom, coin.amount)
            else:
                self._cd[coin.denom] = coin + self._cd[coin.denom]

    def __repr__(self) -> str:
        rstr = ", ".join(f"{c.denom}={c.amount!r}" for c in self.coins)
        return f"Coins({rstr})"

    def __str__(self) -> str:
        return ", ".join(str(coin) for coin in self.coins)

    def to_data(self) -> List[Dict[str, str]]:
        return [coin.to_data() for coin in self.coins]

    def _pretty_repr_(self, path: str = "") -> str:
        d = JiguBox({coin.denom: coin.amount for coin in self.coins})
        return d._pretty_repr_()

    def __add__(self, other: Union[Coin, Coins]) -> Coins:
        if other == 0:
            return Coins(self.coins)
        elif isinstance(other, Coins):
            return Coins(self.coins + other.coins)
        elif isinstance(other, Coin):
            return Coins(self.coins + [other])
        else:
            raise TypeError(
                f"unsupported operand types for +: 'Coins' and '{type(other)}'"
            )

    def __radd__(self, other):
        return self + other

    def __mul__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin * other for coin in self.coins])

    def __rmul__(self, other) -> Coins:
        return self * other

    def __truediv__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin / other for coin in self.coins])

    def __floordiv__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin / other for coin in self.coins])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Coins):
            return JsonSerializable.__eq__(self, other)
        elif isinstance(other, list):
            try:
                return JsonSerializable.__eq__(self, Coins(other))
            except AttributeError:
                return False
        else:
            return False

    @classmethod
    def from_data(cls, data: List[Dict[str, str]]) -> Coins:
        coins = map(Coin.from_data, data)
        return cls(coins)

    @property
    def denoms(self) -> List[str]:
        return sorted(self._cd.keys())

    @property
    def coins(self) -> List[Coin]:
        return sorted(self._cd.values(), key=lambda c: c.denom)

    @property
    def dec_coins(self) -> Coins:
        return Coins(c.dec_coin for c in self.coins)

    @property
    def int_coins(self) -> Coins:
        return Coins(c.int_coin for c in self.coins)

    def filter(self, predicate: Callable[[Coin], bool]) -> Coins:
        return Coins(c for c in self.coins if predicate(c))

    def __iter__(self):
        return iter(self.coins)

    def __contains__(self, denom: str) -> bool:
        return denom in self._cd

    def __getitem__(self, denom: str) -> Coin:
        return self._cd[denom]

    def __getattr__(self, name: str) -> Coin:
        if name in self.denoms:
            return self[name]
        return self.__getattribute__(name)



import pytest

from testtools import LcdRequestTest


class TestStakingApi:
    def test_delegations_request(self, mock_terra, acc_address, val_address):
        acc = acc_address
        val = val_address
        with pytest.raises(TypeError):
            mock_terra.staking.delegations()
        with pytest.raises(TypeError):
            mock_terra.staking.delegation(delegator=acc)
        with pytest.raises(TypeError):
            mock_terra.staking.delegation(validator=val)
        try:
            mock_terra.staking.delegations(delegator=acc)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/delegators/{acc}/delegations"
        try:
            mock_terra.staking.delegations(validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/validators/{val}/delegations"
        try:
            mock_terra.staking.delegations(delegator=acc, validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/delegators/{acc}/delegations/{val}"
        try:
            mock_terra.staking.delegation(delegator=acc, validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/delegators/{acc}/delegations/{val}"

    def test_unbonding_delegations_request(self, mock_terra, acc_address, val_address):
        acc = acc_address
        val = val_address

        with pytest.raises(TypeError):
            mock_terra.staking.unbonding_delegations()
        with pytest.raises(TypeError):
            mock_terra.staking.unbonding_delegation(delegator=acc)
        with pytest.raises(TypeError):
            mock_terra.staking.unbonding_delegation(validator=val)

        try:
            mock_terra.staking.unbonding_delegations(delegator=acc)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/delegators/{acc}/unbonding_delegations"
        try:
            mock_terra.staking.unbonding_delegations(validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/staking/validators/{val}/unbonding_delegations"
        try:
            mock_terra.staking.unbonding_delegations(delegator=acc, validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert (
                request.url == f"/staking/delegators/{acc}/unbonding_delegations/{val}"
            )
        try:
            mock_terra.staking.unbonding_delegation(delegator=acc, validator=val)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert (
                request.url == f"/staking/delegators/{acc}/unbonding_delegations/{val}"
            )

    def test_redelegations_request(self, mock_terra, acc_address, make_val_address):
        acc = acc_address
        val1 = make_val_address()
        val2 = make_val_address()

        try:
            mock_terra.staking.redelegations()
        except LcdRequestTest as r:
            assert r.method == "get"
            assert r.url == "/staking/redelegations"
            assert r.kwargs["params"] == {}

        try:
            mock_terra.staking.redelegations(acc, val1, val2)
        except LcdRequestTest as r:
            assert r.method == "get"
            assert r.url == "/staking/redelegations"
            p = r.kwargs["params"]
            assert p["delegator"] == acc
            assert p["validator_from"] == val1
            assert p["validator_to"] == val2

    def test_bonded_validators_for(self, mock_terra, acc_address, val_address):
        acc = acc_address
        try:
            mock_terra.staking.bonded_validators_for(acc)
        except LcdRequestTest as r:
            assert r.method == "get"
            assert r.url == f"/staking/delegators/{acc}/validators"



__title__ = "jigu"
__description__ = "Python SDK for Terra"
__url__ = "https://jigu.terra.money/"
__version__ = "0.1.4"
__author__ = "Terraform Labs, PTE. LTD"
__author_email__ = "engineering@terra.money"
__license__ = "MIT"
__copyright__ = "Copyright 2020 Terraform Labs, PTE. LTD"



from __future__ import annotations

from typing import Dict, List, Optional, Union

from cached_property import cached_property

from jigu.client.lcd.api import ApiResponse, project
from jigu.client.lcd.api.modules.auth import account_info_type
from jigu.core import (
    AccAddress,
    Coin,
    Coins,
    Delegation,
    Redelegation,
    UnbondingDelegation,
    ValAddress,
    Validator,
)
from jigu.util.validation import validate_val_address


class AccountQuery(object):
    def __init__(self, terra, address: str):
        self.terra = terra
        self.address = AccAddress(address)

    def __str__(self):
        return self.address

    def __repr__(self):
        return f"AccountQuery({self.address!r}) -> {self.terra}"

    @cached_property
    def validator(self):
        return self.terra.validator(self.address.val_address)

    # Auth

    def info(self) -> Union[ApiResponse, account_info_type]:
        return self.terra.auth.acc_info_for(self.address)

    # Bank

    def balance(self, denom: Optional[str] = None) -> Union[ApiResponse, Coin, Coins]:
        res = self.terra.bank.balance_for(self.address)
        return project(res, res[denom] if denom else res)

    # Distribution

    def rewards(
        self, validator: Optional[ValAddress] = None
    ) -> Union[ApiResponse, Dict[ValAddress, Coins], Coins]:
        if validator:
            validator = validate_val_address(validator)
        res = self.terra.distribution.rewards_for(delegator=self.address)
        rewards = res["rewards"]
        return project(res, rewards[validator] if validator else rewards)

    def total_rewards(self, denom: Optional[str] = None) -> Union[ApiResponse, Coins]:
        res = self.terra.distribution.rewards_for(delegator=self.address)
        total = res["total"]
        return project(res, total[denom] if denom else total)

    def withdraw_address(self) -> Union[ApiResponse, AccAddress]:
        return self.terra.distribution.withdraw_address_for(delegator=self.address)

    # Staking

    def delegations(
        self, validator: Optional[ValAddress] = None
    ) -> Union[ApiResponse, List[Delegation]]:
        return self.terra.staking.delegations(
            delegator=self.address, validator=validator
        )

    def unbonding_delegations(
        self, validator: Optional[ValAddress] = None
    ) -> Union[ApiResponse, List[UnbondingDelegation]]:
        return self.terra.staking.unbonding_delegations(
            delegator=self.address, validator=validator
        )

    def redelegations(
        self, validator_src=None, validator_dst=None
    ) -> Union[ApiResponse, List[Redelegation]]:
        return self.terra.staking.redelegations(
            delegator=self.address,
            validator_src=validator_src,
            validator_dst=validator_dst,
        )

    def bonded_validators(self) -> Union[ApiResponse, List[Validator]]:
        return self.terra.staking.bonded_validators_for(delegator=self.address)

    def staking_txs(self):
        return self.terra.staking.staking_txs_for(delegator=self.address)



from __future__ import annotations

from dataclasses import dataclass

from jigu.core import AccAddress, ValAddress
from jigu.core.msg import StdMsg
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address, validate_val_address

__all__ = [
    "MsgModifyWithdrawAddress",
    "MsgWithdrawDelegationReward",
    "MsgWithdrawValidatorCommission",
]


@dataclass
class MsgModifyWithdrawAddress(StdMsg):

    type = "distribution/MsgModifyWithdrawAddress"
    action = "set_withdraw_address"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^distribution/MsgModifyWithdrawAddress\Z"),
        value=S.OBJECT(delegator_address=S.ACC_ADDRESS, withdraw_address=S.ACC_ADDRESS),
    )

    delegator_address: AccAddress
    withdraw_address: AccAddress

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.withdraw_address = validate_acc_address(self.withdraw_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgModifyWithdrawAddress:
        data = data["value"]
        return cls(
            delegator_address=data["delegator_address"],
            withdraw_address=data["withdraw_address"],
        )


@dataclass
class MsgWithdrawDelegationReward(StdMsg):

    type = "distribution/MsgWithdrawDelegationReward"
    action = "withdraw_delegation_reward"

    schema = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^distribution/MsgWithdrawDelegationReward\Z"),
        value=S.OBJECT(
            delegator_address=S.ACC_ADDRESS, validator_address=S.VAL_ADDRESS
        ),
    )

    delegator_address: AccAddress
    validator_address: ValAddress

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.validator_address = validate_val_address(self.validator_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgWithdrawDelegationReward:
        data = data["value"]
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
        )


@dataclass
class MsgWithdrawValidatorCommission(StdMsg):

    type = "distribution/MsgWithdrawValidatorCommission"
    action = "withdraw_validator_commission"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^distribution/MsgWithdrawValidatorCommission\Z"),
        value=S.OBJECT(validator_address=S.VAL_ADDRESS),
    )

    validator_address: ValAddress

    def __post_init__(self):
        self.validator_address = validate_val_address(self.validator_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgWithdrawValidatorCommission:
        data = data["value"]
        return cls(validator_address=data["validator_address"])



"""Low-level Tendermint Node API."""
from typing import List, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.util.serdes import JiguBox

__all__ = ["TendermintApi"]


class TendermintApi(BaseApi):
    def node_info(self) -> Union[ApiResponse, JiguBox]:
        """Get information for the node.
        Returns:
            Node information with the following schema.
        """
        res = self._api_get("/node_info", unwrap=False)
        return project(res, JiguBox(res))

    def syncing(self) -> Union[ApiResponse, bool]:
        """Get whether the node is currently syncing with updates on the blockchain.

        Returns:
            If node is syncing.
        """
        res = self._api_get("/syncing", unwrap=False)
        return project(res, res["syncing"])

    def block(self, height=None) -> Union[ApiResponse, dict]:
        """Get the raw block data at a height on the blockchain.

        Args:
            height (int, optional): block height

        Returns:
            - The block at the provided height. If no height was specified, get the latest
            block at the time of calling.

        """
        if height:
            res = self._api_get(f"/blocks/{height}", unwrap=False)
        else:
            res = self._api_get("/blocks/latest", unwrap=False)
        return project(res, JiguBox(res))

    def validator_set(self, height=None) -> Union[ApiResponse, dict]:
        """Retrieves the current set of validators in the actively validating set.

        ## Arguments
        - **height** `int` *optional*

            block height

        ## Returns
        List of `dict` with the following keys:

        - **address** `ValConsAddress`

            validator's consensus address

        - **pub_key** `ValConsPubKey`

            validator's consensus public key

        - **proposer_priority** `int`
        - **voting_power** `int`
        """
        if height:
            res = self._api_get(f"/validatorsets/{height}")
        else:
            res = self._api_get("/validatorsets/latest")
        vs = res["validators"]
        results = []
        for v in vs:
            results.append(
                JiguBox(
                    {
                        "address": v["address"],
                        "pub_key": v["pub_key"],
                        "proposer_priority": int(v["proposer_priority"]),
                        "voting_power": int(v["voting_power"]),
                    }
                )
            )
        return project(res, results)



import json

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from jigu.core.oracle import ExchangeRateVote
from testtools import assert_serdes_consistent, assert_serdes_exact, load_data


@pytest.fixture
def examples(tdd):
    return load_data(tdd, "objects/oracle/ExchangeRateVote.json")


class TestExchangeRateVoteSerdes:
    def test_schema_valid(self, examples):
        for x in examples:
            assert_serdes_consistent(ExchangeRateVote, x)
            assert_serdes_exact(ExchangeRateVote, x)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(x=from_schema(ExchangeRateVote.__schema__))
    def test_serdes_consistent(self, x):
        assert_serdes_consistent(ExchangeRateVote, x)



import asyncio
import json
from urllib.parse import urljoin

import websockets

from jigu.error import RpcError, get_rpc_error
from jigu.util import hash_amino
from jigu.util.serdes import JiguBox


class WebSocketClient(object):
    def __init__(self, terra, ws_url: str = "", ssl_context=None):
        self.terra = terra
        self.url = ws_url
        self.ssl_context = ssl_context

    def _create_url(self, path):
        return urljoin(self.url, path)

    def connect(self, path="websocket"):
        url = self._create_url(path)
        if self.ssl_context is not None or url.startswith("wss"):
            return websockets.connect(
                self._create_url(path), ssl=self.ssl_context or True
            )
        else:
            return websockets.connect(self._create_url(path))

    @staticmethod
    async def request(ws, method, params={}, id="0"):
        await ws.send(
            json.dumps(
                {"jsonrpc": "2.0", "method": method, "id": id, "params": params,}
            )
        )
        return await ws.recv()

    @staticmethod
    def handle_response(resp: str) -> dict:
        resp = json.loads(resp)
        if "error" in resp:
            raise get_rpc_error(resp["error"]["code"], resp["error"]["data"])
        return resp



import pytest

from jigu.error import InvalidAccAddress
from testtools import LcdRequestTest


class TestBankApi:
    def test_balance_for_request(self, mock_terra, acc_address, val_address):
        address = acc_address
        try:
            mock_terra.bank.balance_for(address)
        except LcdRequestTest as request:
            assert request.method == "get"
            assert request.url == f"/bank/balances/{address}"

        with pytest.raises(InvalidAccAddress):
            mock_terra.bank.balance_for(val_address)



from __future__ import annotations

from typing import List, Optional, Union

from cached_property import cached_property

import jigu.client.object_query
import jigu.client.terra
from jigu.client.lcd.api import ApiResponse
from jigu.core import (
    AccAddress,
    Delegation,
    ExchangeRatePrevote,
    ExchangeRateVote,
    Redelegation,
    UnbondingDelegation,
    ValAddress,
    Validator
)


class ValidatorQuery(object):
    def __init__(self, terra: jigu.client.terra.Terra, val_address: ValAddress):
        self.terra = terra
        self.val_address = ValAddress(val_address)

    def __str__(self):
        return self.val_address

    def __repr__(self):
        return f"ValidatorQuery({self.val_address!r}) -> {self.terra}"

    # Staking
    def info(self) -> Validator:
        return self.terra.staking.val_info_for(self.val_address)

    @cached_property
    def account(self):
        return self.terra.account(self.val_address.acc_address)

    def delegations(
        self, delegator: Optional[AccAddress] = None
    ) -> Union[ApiResponse, List[Delegation]]:
        return self.terra.staking.delegations(
            validator=self.val_address, delegator=delegator
        )

    def unbonding_delegations(
        self, delegator: Optional[AccAddress] = None
    ) -> Union[ApiResponse, List[UnbondingDelegation]]:
        return self.terra.staking.unbonding_delegations(
            validator=self.val_address, delegator=delegator
        )

    def outbound_redelegations(
        self, delegator: AccAddress = None, validator_to: ValAddress = None
    ) -> Union[ApiResponse, List[Redelegation]]:
        return self.terra.staking.redelegations()

    def incoming_redelegations(
        self, delegator: AccAddress = None, validator_from: ValAddress = None
    ) -> Union[ApiResponse, List[Redelegation]]:
        return self.terra.staking.redelegations()

    def val_rewards(self, key: Optional[str] = None):
        return self.terra.distribution.val_rewards_for(self.val_address, key)

    # Oracle
    def feeder(self) -> jigu.client.object_query.AccountQuery:
        return self.terra.account(
            self.terra.oracle.feeder_address_for(validator=self.val_address)
        )

    def misses(self) -> int:
        return self.terra.oracle.misses_for(validator=self.val_address)

    def votes(
        self, denom: Optional[str] = None
    ) -> Union[ApiResponse, List[ExchangeRateVote]]:
        return self.terra.oracle.votes(validator=self.val_address, denom=denom)

    def prevotes(
        self, denom: Optional[str] = None
    ) -> Union[ApiResponse, List[ExchangeRatePrevote]]:
        return self.terra.oracle.prevotes(validator=self.val_address, denom=denom)



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core import Coin, Dec
from jigu.core.treasury import PolicyConstraints
from testtools import assert_serdes_consistent, assert_serdes_exact


@pytest.fixture
def default_tax_policy():
    return PolicyConstraints(
        rate_min=Dec.with_prec(5, 4),
        rate_max=Dec.with_prec(1, 2),
        cap=Coin("usdr", 1),
        change_max=Dec.with_prec(25, 5),
    )


@pytest.fixture
def default_tax_rate():
    return Dec.with_prec(1, 3)


class TestPolicyConstraints:
    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(PolicyConstraints.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(PolicyConstraints, m)

    def test_clamp(self, default_tax_policy, default_tax_rate):
        # Analogous test:
        # https://github.com/terra-project/core/blob/develop/x/treasury/internal/types/constraint_test.go#L10
        tax_policy = default_tax_policy
        prev_rate = default_tax_rate

        # Case 1: try to update delta > change_max
        new_rate = prev_rate + tax_policy.change_max * 2
        clamped_rate = tax_policy.clamp(prev_rate, new_rate)
        assert prev_rate + tax_policy.change_max == clamped_rate

        # Case 2: try to update delta > change_max (negative)
        new_rate = prev_rate - tax_policy.change_max * 2
        clamped_rate = tax_policy.clamp(prev_rate, new_rate)
        assert prev_rate - tax_policy.change_max == clamped_rate

        # Case 3: try to update the new rate > min_rate
        prev_rate = tax_policy.rate_max
        new_rate = prev_rate + Dec.with_prec(1, 3)
        clamped_rate = tax_policy.clamp(prev_rate, new_rate)
        assert tax_policy.rate_max == clamped_rate

        # Case 4: try to update the new rate < minRate
        prev_rate = tax_policy.rate_min
        new_rate = tax_policy.rate_min - Dec.with_prec(1, 3)
        clamped_rate = tax_policy.clamp(prev_rate, new_rate)
        assert tax_policy.rate_min == clamped_rate



import warnings
from typing import Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import AccAddress, Account, LazyGradedVestingAccount
from jigu.error import AccountNotFoundWarning
from jigu.util.validation import validate_acc_address

__all__ = ["account_info_type", "AuthApi"]
account_info_type = Union[Account, LazyGradedVestingAccount]


class AuthApi(BaseApi):
    def acc_info_for(
        self, address: AccAddress
    ) -> Union[ApiResponse, account_info_type]:
        address = validate_acc_address(address)
        info = self._api_get(f"/auth/accounts/{address}")
        if info["type"] == "core/Account":
            res = Account.from_data(info)
        elif info["type"] == "core/LazyGradedVestingAccount":
            res = LazyGradedVestingAccount.from_data(info)
        else:
            raise ValueError("could not deserialize account in auth.acc_info")
        if res.address is None:
            warnings.warn(
                "Account was not found; perhaps wrong chain or account needs to first be sent funds.",
                AccountNotFoundWarning,
            )
        return project(info, res)



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from jigu.core.staking.validator import Validator
from testtools import assert_serdes_consistent, assert_serdes_exact, load_data


@pytest.fixture(scope="module")
def examples(tdd):
    return load_data(tdd, "objects/staking/Validator.json")


class TestValidatorSerdes:
    def test_schema_valid(self, examples):
        for v in examples:
            assert_serdes_consistent(Validator, v)
            assert_serdes_exact(Validator, v)

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(v=from_schema(Validator.__schema__))
    def test_serdes_consistent(self, v):
        assert_serdes_consistent(Validator, v)



from __future__ import annotations

from typing import Iterable, Union

from jigu.core.auth.transaction import TxInfo
from jigu.query.msginfo import MsgInfosQuery
from jigu.util.pretty import PrettyPrintable, pretty_repr

__all__ = ["TxInfosQuery"]


class TxInfosQuery(PrettyPrintable):
    def __init__(self, txinfos: Iterable[TxInfo]):
        self.txinfos = list(txinfos)

    def __repr__(self) -> str:
        return f"<TxInfosQuery {len(self.txinfos)} txs, {len(self.msgs)} msgs>"

    def __getitem__(self, key) -> Union[TxInfo, TxInfosQuery]:
        if isinstance(key, int):
            return self.txinfos[key]
        elif isinstance(key, slice):
            return TxInfosQuery(self.txinfos[key])
        else:
            raise KeyError(
                f"unusuable key {key} for TxInfosQuery (use 'int' or 'slice')"
            )

    def __len__(self) -> int:
        return len(self.txinfos)

    @property
    def msgs(self) -> MsgInfosQuery:
        return MsgInfosQuery([msg for t in self.txinfos for msg in t.msgs])

    @property
    def pretty_iterable(self) -> bool:
        return True

    def _pretty_repr_(self, path: str = "") -> str:
        return pretty_repr(self.txinfos, path)



"""Jigu 지구, the Python SDK for Terra."""

# Set default logging to avoid NoHandler warnings
import logging

from .__version__ import *
from .client.terra import Terra

logging.getLogger(__name__).addHandler(logging.NullHandler())



import os
import sys

import pytest
import rstr

from jigu import Terra
from jigu.core.auth.transaction import StdFee
from jigu.key.mnemonic import MnemonicKey

sys.path.append(os.path.join(os.path.dirname(__file__), "helpers"))  # isort:skip


# Set up testing environment
os.environ["JIGU_BECH32_VALIDATE"] = "0"

pytest.register_assert_rewrite("testtools")
from testtools import (  # isort:skip
    load_data,
    lcd_request_test_middleware,
)

# Config


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


# Fixtures


@pytest.fixture(scope="session")
def tdd():  # short for test data dir
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


@pytest.fixture
def acc_address():
    return rstr.xeger(r"^terra[0-9a-z]{39}\Z")


@pytest.fixture
def make_acc_address():
    return lambda: rstr.xeger(r"^terra[0-9a-z]{39}\Z")


@pytest.fixture
def val_address():
    return rstr.xeger(r"^terravaloper[0-9a-z]{39}\Z")


@pytest.fixture
def make_val_address():
    return lambda: rstr.xeger(r"^terravaloper[0-9a-z]{39}\Z")


@pytest.fixture
def mnemonics(tdd):
    """Loads a series of known mnemonics and their generated addresses on account 0, index 0."""
    return load_data(tdd, "mnemonics.json")


@pytest.fixture(autouse=True, scope="session")
def mock_terra():
    """Mocks the Terra instance before a request is made."""
    terra = Terra("soju-0013", "")
    terra.lcd.request_middlewares.append(lcd_request_test_middleware)
    return terra


@pytest.fixture
def terra():
    """Testnet Terra connection."""
    return Terra("soju-0013", "https://soju-lcd.terra.dev")


@pytest.fixture
def wallet(mnemonics, terra):
    """Wallet #1, which should have some testnet funds."""
    m = mnemonics[0]["mnemonic"]
    return terra.wallet(MnemonicKey(m))


@pytest.fixture
def fee():
    return StdFee.make(gas=1000000, uluna=1000000)  # to make sure tx passes



"""Serialization of events.
NOTE: Due to lack of consistency throughout Cosmos-SDK and Terra-Core codebases
regarding events, the serialization is very basic and mostly schema-less.

TODO: Make proper event handling middlewares accessible from terra.tx.transformer
"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import List

from jigu.util.serdes import JiguBox, JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = ["Event"]

# TODO: Properly treat events in terra-core codebase.

# Example of what event code might look like.
#
# @dataclass
# class MessageEvent:
#
#     type = "message"
#     module: str
#     action: str
#     sender: List[str]
#
#     def __post_init__(self):
#         self.module = self.module[0]
#         self.action = self.action[0]
#
#
# @dataclass
# class TransferEvent:
#
#     type = "transfer"
#     amount: List[Coin]
#     recipient: List[str]
#
#     def __post_init__(self):
#         self.amount = [Coin.from_str(c) for c in self.amount]
#
#
# @dataclass
# class SwapEvent:
#
#     type = "swap"
#     offer: Coin
#     swap_coin: Coin
#     swap_fee: Coin
#     trader: str
#
#     def __post_init__(self):
#         self.offer = Coin.from_str(self.offer[0])
#         self.swap_coin = Coin.from_str(self.swap_coin[0])
#         self.swap_fee = Coin.from_str(self.swap_fee[0])
#         self.trader = self.trader[0]
#
#
# EVENT_TYPES = {
#     "message": MessageEvent,
#     "transfer": TransferEvent,
#     "swap": SwapEvent,
# }


@dataclass
class Event(JsonSerializable, JsonDeserializable):
    __schema__ = S.OBJECT(
        type=S.STRING, attributes=S.ARRAY(S.OBJECT(key=S.STRING, value=S.STRING))
    )

    type: str
    attributes: JiguBox[str, List[str]]

    def __repr__(self) -> str:
        return f"<Event {self.type}>"

    def __getitem__(self, item) -> List[str]:
        return self.attributes[item]

    def __getattr__(self, name) -> List[str]:
        if name in self.attributes:
            return self.attributes[name]
        return self.__getattribute__(name)

    @property
    def pretty_data(self):
        d = {"type": self.type, **self.attributes.to_data()}
        return d.items()

    def to_data(self):
        return {
            "type": self.type,
            "attributes": [
                {"key": k, "value": v} for k, vs in self.attributes.items() for v in vs
            ],
        }

    @classmethod
    def from_data(cls, data: dict) -> Event:
        attrs = defaultdict(list)
        for attr in data["attributes"]:
            attrs[attr["key"]].append(attr["value"])
        return cls(type=data["type"], attributes=JiguBox(attrs))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from jigu.core import AccAddress, Coin, Dec, ValAddress, ValConsPubKey
from jigu.core.msg import StdMsg
from jigu.core.staking import CommissionRates, Description
from jigu.util.validation import (
    Schemas as S,
    validate_acc_address,
    validate_val_address,
    validate_val_conspubkey,
)


@dataclass
class MsgBeginRedelegate(StdMsg):

    type = "staking/MsgBeginRedelegate"
    action = "begin_redelegate"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^staking/MsgBeginRedelegate\Z"),
        value=S.OBJECT(
            delegator_address=S.ACC_ADDRESS,
            validator_src_address=S.VAL_ADDRESS,
            validator_dst_address=S.VAL_ADDRESS,
            amount=Coin.__schema__,
        ),
    )

    delegator_address: AccAddress
    validator_src_address: ValAddress
    validator_dst_address: ValAddress
    amount: Coin

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.validator_src_address = validate_val_address(self.validator_src_address)
        self.validator_dst_address = validate_val_address(self.validator_dst_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgBeginRedelegate:
        data = data["value"]
        return cls(
            delegator_address=data["delegator_address"],
            validator_src_address=data["validator_src_address"],
            validator_dst_address=data["validator_dst_address"],
            amount=Coin.from_data(data["amount"]),
        )


@dataclass
class MsgDelegate(StdMsg):

    type = "staking/MsgDelegate"
    action = "delegate"

    schema = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^staking/MsgDelegate\Z"),
        value=S.OBJECT(
            delegator_address=S.ACC_ADDRESS,
            validator_address=S.VAL_ADDRESS,
            amount=Coin.__schema__,
        ),
    )

    delegator_address: AccAddress
    validator_address: ValAddress
    amount: Coin

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.validator_address = validate_val_address(self.validator_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgDelegate:
        data = data["value"]
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            amount=Coin.from_data(data["amount"]),
        )


@dataclass
class MsgUndelegate(StdMsg):

    type = "staking/MsgUndelegate"
    action = "begin_unbonding"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^staking/MsgUndelegate\Z"),
        value=S.OBJECT(
            delegator_address=S.ACC_ADDRESS,
            validator_address=S.VAL_ADDRESS,
            amount=Coin.__schema__,
        ),
    )

    delegator_address: AccAddress
    validator_address: ValAddress
    amount: Coin

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.validator_address = validate_val_address(self.validator_address)

    @classmethod
    def from_data(cls, data: dict) -> MsgUndelegate:
        data = data["value"]
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            amount=Coin.from_data(data["amount"]),
        )


@dataclass
class MsgEditValidator(StdMsg):

    type = "staking/MsgEditValidator"
    action = "edit_validator"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^staking/MsgEditValidator\Z"),
        value=S.OBJECT(
            Description=Description.__schema__,
            address=S.VAL_ADDRESS,
            commission_rate=S.OPTIONAL(Dec.__schema__),
            min_self_delegation=S.OPTIONAL(S.STRING_INTEGER),
        ),
    )

    Description: Description
    address: ValAddress
    commission_rate: Optional[Dec] = None
    min_self_delegation: Optional[int] = None

    def __post_init__(self):
        if self.commission_rate is not None:
            self.commission_rate = Dec(self.commission_rate)
        if self.min_self_delegation is not None:
            self.min_self_delegation = int(self.min_self_delegation)
        self.address = validate_val_address(self.address)

    def msg_value(self) -> dict:
        d = dict(self.__dict__)
        msd = self.min_self_delegation
        if msd is not None:
            msd = str(msd)
        d["min_self_delegation"] = msd
        return d

    @classmethod
    def from_data(cls, data: dict) -> MsgEditValidator:
        data = data["value"]
        msd = int(data["min_self_delegation"]) if data["min_self_delegation"] else None
        cr = Dec(data["commission_rate"]) if data["commission_rate"] else None
        return cls(
            Description=data["Description"],
            address=data["address"],
            commission_rate=cr,
            min_self_delegation=msd,
        )


@dataclass
class MsgCreateValidator(StdMsg):

    type = "staking/MsgCreateValidator"
    action = "create_validator"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^staking/MsgCreateValidator\Z"),
        value=S.OBJECT(
            description=Description.__schema__,
            commission=CommissionRates.__schema__,
            min_self_delegation=S.STRING_INTEGER,
            delegator_address=S.ACC_ADDRESS,
            validator_address=S.VAL_ADDRESS,
            pubkey=S.VAL_CONSPUBKEY,
            value=Coin.__schema__,
        ),
    )

    description: Description
    commission: CommissionRates
    min_self_delegation: int
    delegator_address: AccAddress
    validator_address: ValAddress
    pubkey: ValConsPubKey
    value: Coin

    def __post_init__(self):
        self.delegator_address = validate_acc_address(self.delegator_address)
        self.validator_address = validate_val_address(self.validator_address)
        validate_val_conspubkey(self.pubkey)
        self.min_self_delegation = int(self.min_self_delegation)

    def msg_value(self) -> dict:
        d = dict(self.__dict__)
        msd = self.min_self_delegation
        d["min_self_delegation"] = str(msd)
        return d

    @classmethod
    def from_data(cls, data: dict) -> MsgCreateValidator:
        data = data["value"]
        return cls(
            description=Description.from_data(data["description"]),
            commission=CommissionRates.from_data(data["commission"]),
            min_self_delegation=int(data["min_self_delegation"]),
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            pubkey=data["pubkey"],
            value=Coin.from_data(data["value"]),
        )



from __future__ import annotations

import abc
from typing import Any, Dict

from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

# __all__ exports all messages


class StdMsg(JsonSerializable, JsonDeserializable, metaclass=abc.ABCMeta):

    __schema__ = S.OBJECT(type=S.STRING, value={"type": "object"})

    @property
    @abc.abstractmethod
    def type(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def action(self):
        raise NotImplementedError

    def msg_value(self) -> Dict[str, Any]:
        return dict(self.__dict__)

    def to_data(self) -> Dict[str, Any]:
        return {"type": self.type, "value": self.msg_value()}


from .bank import MsgSend, MsgMultiSend  # isort:skip
from .distribution import (  # isort:skip
    MsgModifyWithdrawAddress,
    MsgWithdrawDelegationReward,
    MsgWithdrawValidatorCommission,
)
from .gov import MsgSubmitProposal, MsgVote, MsgDeposit  # isort:skip
from .market import MsgSwap  # isort:skip
from .staking import (  # isort:skip
    MsgBeginRedelegate,
    MsgDelegate,
    MsgUndelegate,
    MsgEditValidator,
    MsgCreateValidator,
)
from .slashing import MsgUnjail  # isort:skip
from .oracle import (  # isort:skip
    MsgDelegateFeedConsent,
    MsgExchangeRatePrevote,
    MsgExchangeRateVote,
)

MSG_TYPES = {
    "bank/MsgSend": MsgSend,
    "bank/MsgMultiSend": MsgMultiSend,
    "distribution/MsgModifyWithdrawAddress": MsgModifyWithdrawAddress,
    "distribution/MsgWithdrawDelegationReward": MsgWithdrawDelegationReward,
    "distribution/MsgWithdrawValidatorCommission": MsgWithdrawValidatorCommission,
    "oracle/MsgExchangeRatePrevote": MsgExchangeRatePrevote,
    "oracle/MsgExchangeRateVote": MsgExchangeRateVote,
    "oracle/MsgDelegateFeedConsent": MsgDelegateFeedConsent,
    "gov/MsgSubmitProposal": MsgSubmitProposal,
    "gov/MsgDeposit": MsgDeposit,
    "gov/MsgVote": MsgVote,
    "market/MsgSwap": MsgSwap,
    "staking/MsgBeginRedelegate": MsgBeginRedelegate,
    "staking/MsgDelegate": MsgDelegate,
    "staking/MsgUndelegate": MsgUndelegate,
    "staking/MsgEditValidator": MsgEditValidator,
    "staking/MsgCreateValidator": MsgCreateValidator,
    "cosmos/MsgUnjail": MsgUnjail,
}



"""Tools for building WebSocket-based listeners."""
import asyncio
import base64
from typing import Any, Callable, List

from jigu.core import TxInfo
from jigu.core.event import Event
from jigu.error import RpcError
from jigu.query.event import EventsQuery
from jigu.util import hash_amino
from jigu.util.serdes import JiguBox

__all__ = [
    "Ignore",
    "StopListening",
    "Unsubscribe",
    "Resubscribe",
    "TendermintEventListener",
    "TxListener",
    "BlockListener",
    "run_listeners",
    "Q",
]


class Ignore(Exception):
    """Raise to ignore the error."""


class StopListening(Exception):
    """Raise to stop listening."""


class Unsubscribe(Exception):
    """Raise to unsubscribe."""


class Resubscribe(Exception):
    """Raise to resubscribe."""


class TendermintEventListener(object):

    event = ""
    query = {}

    def __init__(
        self, terra, *, func: Callable[[TxInfo], Any] = None, query: dict = None
    ):
        self.terra = terra
        self.terra = terra
        self.func = func
        if query is not None:
            self.query = query  # otherwise, it is left as default ({})
        self.ws = None

    def on_message(self, msg):
        """Override this if you want access to the direct JSON-RPC parsed message. For
        implementing custom listeners."""
        pass

    def on_error(self, err: RpcError):
        """Implement `on_error`, which will be provided the RpcError, if encountered.
        Note that any other errors will not make it here, and should be handled in `on_tx`.

        With regard to execution, you can do one of four things:

        1) ignore the error

        >>> self.ignore()

        2) stop listening (stop execution)

        >>> self.stop_listening()

        3) unsubscribe

        >>> self.unsubscribe()

        4) resubscribe

        >>> self.resubscribe()

        If you do nothing, the error will be raised outside listening execution context.
        """
        pass

    async def _send_subscription_msg(self, tm_event, query, unsubscribe=False):
        # TODO: improve query ... can have stupid injection mistakes
        if not query:  # if query is not truthy; i.e. {}, '', None, ...
            query = ""
        else:
            if isinstance(query, dict):
                query = " AND ".join(f"{i[0]} = '{i[1]}'" for i in query.items())
            elif hasattr(query, "to_tmquery"):
                query = query.to_tmquery()
            else:  # use the string
                query = str(query)
            query = " AND " + query
        tm_query = f"tm.event = '{tm_event}'"
        sub_type = "subscribe" if not unsubscribe else "unsubscribe"
        resp = await self.terra.ws.request(
            self.ws, sub_type, params={"query": f"{tm_query}{query}"}
        )
        return self.terra.ws.handle_response(resp)

    async def _subscribe(self):
        """Performs the actual subscribing."""
        await self._send_subscription_msg(self.event, self.query)

    async def _unsubscribe(self):
        """Performs the actual unsubscribing."""
        await self._send_subscription_msg(self.event, self.query, unsubscribe=True)

    # error handling methods (these must be called inside `on_error`).

    def stop_listening(self):
        """Used inside `on_error`, tells the listener to stop listening."""
        raise StopListening

    def ignore(self):
        """Used inside `on_error`, tells the listener to ignore the error."""
        raise Ignore

    def resubscribe(self):
        """Used inside `on_error`, tells the listener to re-subscribe."""
        raise Resubscribe

    def unsubscribe(self):
        """Used inside `on_error`, tells the listener to unsubscribe."""
        raise Unsubscribe

    def __await__(self):
        async def listener():
            async with self.terra.ws.connect() as ws:
                self.ws = ws
                await self._subscribe()
                while True:
                    try:  # for stop listening
                        try:
                            async for resp in ws:
                                resp = self.terra.ws.handle_response(resp)
                                result = resp["result"]
                                self.on_message(result)
                        except RpcError as err:
                            try:
                                self.on_error(err)
                            except Ignore:
                                continue
                            except Unsubscribe:
                                await self._unsubscribe()
                                continue
                            except Resubscribe:
                                await self._unsubscribe()
                                await self._subscribe()
                                continue
                            raise err
                    except StopListening:
                        return

        return listener()


class TxListener(TendermintEventListener):

    event = "Tx"
    query = {}

    def _process_tx_msg(self, message):
        """Currently, the default behavior is to re-lookup the tranasaction via the LCD
        connection, because the result is amino-encoded. Follow this example to build
        your own version by overriding the default behavior in _process_tx_msg."""

        txresult = message["data"]["value"]["TxResult"]
        return self.terra.tx_info(hash_amino(txresult["tx"]))

    def on_message(self, message):
        tx = self._process_tx_msg(message)
        return self.on_tx(tx)

    def on_tx(self, tx: TxInfo):
        """Implement `on_tx`, which will be provided the transaction information received
        through the subscription.
        """
        if self.func:
            self.func(self, tx)


class BlockListener(TendermintEventListener):

    event = "NewBlock"
    query = {}

    @staticmethod
    def _decode_kvs_b64(atts: List[dict]):
        """Decodes base64 key-value pairs for events."""
        return [
            {
                "key": base64.b64decode(att["key"]).decode(),
                "value": base64.b64decode(att["value"]).decode(),
            }
            for att in atts
        ]

    def _process_block_msg(self, message) -> JiguBox:
        """Will process the response from a NewBlock subscription and turn it into a
        block's context."""
        data = message["data"]["value"]
        height = data["block"]["header"]["height"]
        block = self.terra.blocks.at(height)
        begin_block_events = [
            {"type": e["type"], "attributes": self._decode_kvs_b64(e["attributes"])}
            for e in data["result_begin_block"].get("events", [])
        ]
        end_block_events = [
            {"type": e["type"], "attributes": self._decode_kvs_b64(e["attributes"])}
            for e in data["result_end_block"].get("events", [])
        ]
        return JiguBox(
            {
                "block": block,
                "events": {
                    "begin_block": EventsQuery(
                        Event.deserialize(e) for e in begin_block_events
                    ),
                    "end_block": EventsQuery(
                        Event.deserialize(e) for e in end_block_events
                    ),
                },
            }
        )

    def on_message(self, message):
        block = self._process_block_msg(message)
        return self.on_block(block)

    def on_block(self, block_ctx: JiguBox):
        """Implement `on_block`, which will be provided the block information received
        through the subscription.
        """
        if self.func:
            self.func(self, block_ctx)


def run_listeners(*listeners):
    """Runs all listeners until they all stop."""
    asyncio.get_event_loop().run_until_complete(asyncio.gather(*listeners))


# NOTE: The following is experimental.


class QueryBuilder:
    def __init__(self, path):
        self.path = path

    def eq(self, arg):
        return QueryNode("=", self.path, arg)

    def ne(self, arg):
        raise NotImplementedError(
            'operation can be "=", "<", "<=", ">", ">=", "CONTAINS" AND "EXISTS"'
        )

    def lt(self, arg):
        return QueryNode("<", self.path, arg)

    def le(self, arg):
        return QueryNode("<=", self.path, arg)

    def gt(self, arg):
        return QueryNode(">", self.path, arg)

    def ge(self, arg):
        return QueryNode(">=", self.path, arg)

    def __contains__(self, arg):
        return QueryNode("CONTAINS", self.path, arg)

    def contains(self, arg):
        return self.__contains__(arg)

    @property
    def exists(self):
        return ExistsNode(self.path)


class QueryNode:
    def __init__(self, op, path, arg):
        self.op = op
        self.path = path
        self.arg = arg

    def __and__(self, other):
        return AndNode(self, other)

    def to_tmquery(self):
        return f"{self.path} {self.op} {self.arg!r}"


class AndNode(QueryNode):
    def __init__(self, t1: QueryNode, t2: QueryNode):
        self.t1 = t1
        self.t2 = t2

    def to_tmquery(self):
        return f"{self.t1.to_tmquery()} AND {self.t2.to_tmquery()}"


class ExistsNode(QueryNode):
    def __init__(self, path):
        self.path = path

    def to_tmquery(self):
        return f"{self.path} EXISTS"


# Usage: Q("message.action").exists & Q("message.action").eq("3")
Q = QueryBuilder



from __future__ import annotations

import hashlib
from dataclasses import dataclass
from decimal import Decimal
from typing import Type, Union

from jigu.core import AccAddress, Coin, Dec, ValAddress
from jigu.core.msg import StdMsg
from jigu.util.validation import Schemas as S
from jigu.util.validation import (
    validate_acc_address,
    validate_same_denom,
    validate_val_address,
)

__all__ = [
    "vote_hash",
    "MsgExchangeRatePrevote",
    "MsgExchangeRateVote",
    "MsgDelegateFeedConsent",
]


def vote_hash(salt: str, exchange_rate: Dec, denom: str, validator: str) -> str:
    payload = f"{salt}:{exchange_rate}:{denom}:{validator}"
    sha_hash = hashlib.sha256(payload.encode())
    return sha_hash.hexdigest()[:40]


@dataclass
class MsgExchangeRatePrevote(StdMsg):

    type = "oracle/MsgExchangeRatePrevote"
    action = "exchangerateprevote"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^oracle/MsgExchangeRatePrevote\Z"),
        value=S.OBJECT(
            hash=S.STRING,
            denom=S.STRING,
            feeder=S.ACC_ADDRESS,
            validator=S.VAL_ADDRESS,
        ),
    )

    hash: str
    denom: str
    feeder: AccAddress
    validator: ValAddress

    def __post_init__(self):
        self.feeder = AccAddress(self.feeder)
        self.validator = ValAddress(self.validator)

    @classmethod
    def from_data(cls, data: dict) -> MsgExchangeRatePrevote:
        data = data["value"]
        return cls(
            hash=data["hash"],
            denom=data["denom"],
            feeder=data["feeder"],
            validator=data["validator"],
        )


@dataclass
class MsgExchangeRateVote(StdMsg):

    type = "oracle/MsgExchangeRateVote"
    action = "exchangeratevote"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^oracle/MsgExchangeRateVote\Z"),
        value=S.OBJECT(
            exchange_rate=Dec.__schema__,
            salt=S.STRING,
            denom=S.STRING,
            feeder=S.ACC_ADDRESS,
            validator=S.VAL_ADDRESS,
        ),
    )

    exchange_rate: Union[str, Type[Decimal], Coin, int]
    salt: str
    denom: str
    feeder: AccAddress
    validator: ValAddress

    def __post_init__(self):
        self.feeder = validate_acc_address(self.feeder)
        self.validator = validate_val_address(self.validator)
        if not isinstance(self.exchange_rate, Coin):
            self.exchange_rate = Coin(self.denom, self.exchange_rate)
        else:
            validate_same_denom(self.exchange_rate.denom, self.denom)

    def msg_value(self) -> dict:
        d = dict(self.__dict__)
        d["exchange_rate"] = str(self.exchange_rate.amount)
        return d

    @property
    def vote_hash(self):
        return vote_hash(
            self.salt, self.exchange_rate.amount, self.denom, self.validator
        )

    @property
    def prevote(self):
        return MsgExchangeRatePrevote(
            hash=self.vote_hash,
            denom=self.denom,
            feeder=self.feeder,
            validator=self.validator,
        )

    @classmethod
    def from_data(cls, data: dict) -> MsgExchangeRateVote:
        data = data["value"]
        xr = Coin(data["denom"], data["exchange_rate"])
        return cls(
            exchange_rate=xr,
            salt=data["salt"],
            denom=xr.denom,
            feeder=data["feeder"],
            validator=data["validator"],
        )


@dataclass
class MsgDelegateFeedConsent(StdMsg):

    type = "oracle/MsgDelegateFeedConsent"
    action = "delegatefeeder"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^oracle/MsgDelegateFeedConsent\Z"),
        value=S.OBJECT(operator=S.VAL_ADDRESS, delegate=S.ACC_ADDRESS),
    )

    operator: ValAddress
    delegate: AccAddress

    def __post_init__(self):
        self.operator = validate_val_address(self.operator)
        self.delegate = validate_acc_address(self.delegate)

    @classmethod
    def from_data(cls, data: dict) -> MsgDelegateFeedConsent:
        data = data["value"]
        return cls(operator=data["operator"], delegate=data["delegate"])



from __future__ import annotations

import hashlib

from ecdsa import SECP256k1, SigningKey
from ecdsa.util import sigencode_string_canonize
from mnemonic import Mnemonic

from jigu.key import Key, derive_child, derive_root, LUNA_COIN_TYPE

__all__ = ["MnemonicKey"]

# TODO: add test for coin_type


class MnemonicKey(Key):
    """Implements Key interface with 24-word mnemonic. This implementation exposes the private
    key inside the application, and may not be suited for situations that demand high security."""

    def __init__(
        self,
        mnemonic: str,
        account: int = 0,
        index: int = 0,
        coin_type: int = LUNA_COIN_TYPE,
    ):
        self.mnemonic = mnemonic
        seed = Mnemonic("english").to_seed(self.mnemonic)
        root = derive_root(seed)
        child = derive_child(root, account, index, coin_type)
        self.account = account
        self.index = index
        self._private_key = child.PrivateKey()
        self._public_key = child.PublicKey()

    @property
    def public_key(self) -> bytes:
        return self._public_key

    @classmethod
    def generate(
        cls, account: int = 0, index: int = 0, coin_type: int = LUNA_COIN_TYPE
    ) -> MnemonicKey:
        """Generate a random Mnemonic, with the private / public key pair with the BIP44 HD Path
        located at 44'/330'/<account>'/0/<index> (LUNA)."""
        return cls(
            mnemonic=Mnemonic("english").generate(256),
            account=account,
            index=index,
            coin_type=coin_type,
        )

    def sign(self, payload: bytes) -> bytes:
        """Sign a payload using ECDSA curve SECP256k1."""
        sk = SigningKey.from_string(self._private_key, curve=SECP256k1)
        return sk.sign_deterministic(
            payload, hashfunc=hashlib.sha256, sigencode=sigencode_string_canonize,
        )



from __future__ import annotations

import re
from datetime import datetime
from decimal import Decimal

import wrapt

from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = ["Timestamp"]


class Timestamp(wrapt.ObjectProxy, JsonSerializable, JsonDeserializable):

    # we have 2 patterns, 1 for serdes-checking in tests and 1 for checking if it works
    # python's builtin datetime
    pattern = re.compile(
        r"^(0001|[1-9]\d{3})-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])T([0-1]\d|2[0-3]):([0-5]\d):([0-5]\d)(\.\d+)?Z\Z"
    )

    __schema__ = S.STRING_WITH_PATTERN(
        r"^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(\.\d+)?Z\Z"
    )

    def __init__(self, year, month, day, hour, minute, second, ns=0, precision=0):
        wrapt.ObjectProxy.__init__(
            self, datetime(year, month, day, hour, minute, second, ns // 1000)
        )
        self._self_ns = ns
        self._self_precision = min(9, precision)

    @property
    def ns(self):
        return self._self_ns

    @property
    def precision(self):
        return self._self_precision

    @property
    def nanosecond(self):
        return int(self.ns % 1000)

    def __str__(self) -> str:
        Y = str(self.year).rjust(4, "0")
        M = str(self.month).rjust(2, "0")
        D = str(self.day).rjust(2, "0")
        h = str(self.hour).rjust(2, "0")
        m = str(self.minute).rjust(2, "0")
        s = str(self.second).rjust(2, "0")
        r = f"{Y}-{M}-{D}T{h}:{m}:{s}"
        if self.precision > 0:
            r += "."
            r += str(self.microsecond * 1000 + self.nanosecond).rjust(9, "0")[
                : self.precision
            ]
        return r + "Z"

    def __repr__(self) -> str:
        s = self
        p = f", {s.precision}" if s.precision != 0 else ""
        return f"Timestamp({s.year}, {s.month}, {s.day}, {s.hour}, {s.minute}, {s.second}, {s.ns}{p})"

    def to_data(self) -> str:
        return str(self)

    def _pretty_repr_(self, path: str = ""):
        return self.strftime("%a %b %d '%y - %H:%M:%S.%f")

    @classmethod
    def from_datetime(cls, dt: datetime) -> Timestamp:
        return cls(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            ns=dt.microsecond * 1000,
            precision=6 if dt.microsecond > 0 else 0,
        )

    @classmethod
    def from_data(cls, data: str) -> Timestamp:
        m = cls.pattern.match(data)
        # m = re.match(cls.__schema__["pattern"], data)
        if m:
            year, month, day, hour, minute, second = [
                int(m.group(x)) for x in range(1, 7)
            ]
            if m.group(7) is not None:
                # keep only nanosecond precision
                ns = int(Decimal(m.group(7)) * 1_000_000_000)
                precision = min(9, len(m.group(7)) - 1)
            else:
                ns = 0
                precision = 0
            return cls(year, month, day, hour, minute, second, ns, precision)
        else:
            raise ValueError(f"Timestamp could not deserialize: {data}")

    @classmethod
    def from_str(cls, string: str) -> Timestamp:
        return cls.from_data(string)

    # TODO: write datetime analogous functions for things that return datetime
    # Not sure if this is too high priority as Timestamp is not a replacement for datetime.

    def __eq__(self, other):
        if isinstance(other, Timestamp):
            return self.__wrapped__ == other.__wrapped__ and self.ns == other.ns
        return self.__wrapped__ == other and self.nanosecond == 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, Timestamp):
            if self.__wrapped__ == other.__wrapped__:
                return self.ns < other.ns
            else:
                return self.__wrapped__ < other.__wrapped__
        return self.__wrapped__ < other

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        if isinstance(other, Timestamp):
            if self.__wrapped__ == other.__wrapped__:
                return self.ns > other.ns
            else:
                return self.__wrapped__ > other.__wrapped__
        if self.__wrapped__ == other:
            return self.ns > 0
        else:
            return self.__wrapped__ > other

    def __ge__(self, other):
        return self == other or self > other

    # def __cmp__(self, other):
    #     if self < other:
    #         return -1
    #     elif self > other:
    #         return 1
    #     else:
    #         return 0



from typing import Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import Coins

__all__ = ["SupplyApi"]


class SupplyApi(BaseApi):
    def total(self) -> Union[ApiResponse, Coins]:
        res = self._api_get("/supply/total")
        return project(res, Coins.deserialize(res))



from __future__ import annotations

import base64
import hashlib
import uuid

from bech32 import bech32_encode, convertbits

__all__ = ["get_bech", "generate_salt", "hash_amino"]


def get_bech(prefix: str, payload: str) -> str:
    return bech32_encode(
        prefix, convertbits(bytes.fromhex(payload), 8, 5)
    )  # base64 -> base32


def generate_salt() -> str:
    """Generate a 4 bytes salt."""
    return uuid.uuid4().hex[:4]


def hash_amino(txdata: str) -> str:
    """Get the transaction hash from Amino-encoded Transaction in base64."""
    return hashlib.sha256(base64.b64decode(txdata)).digest().hex()



from .auth import AuthApi
from .bank import BankApi
from .distribution import DistributionApi
from .gov import GovApi
from .market import MarketApi
from .oracle import OracleApi
from .slashing import SlashingApi
from .staking import StakingApi
from .supply import SupplyApi
from .treasury import TreasuryApi

__all__ = [
    "AuthApi",
    "BankApi",
    "DistributionApi",
    "GovApi",
    "MarketApi",
    "OracleApi",
    "SlashingApi",
    "StakingApi",
    "SupplyApi",
    "TreasuryApi",
]



import pytest

from jigu import Terra
from jigu.key.mnemonic import MnemonicKey
from jigu.core.msg import *
from jigu.core import StdFee, Coin, Dec, Coins
from jigu.error import CodespaceError, TxCodespaceError
from jigu.core.proposal import *
from jigu.core.treasury import PolicyConstraints


class TestTx:
    def test_broadcast(self, wallet, fee):
        swap = MsgSwap(
            trader=wallet.address, offer_coin=Coin("uluna", 1), ask_denom="umnt"
        )
        tx = wallet.create_and_sign_tx(swap, fee=fee)
        res = wallet.broadcast(tx)
        assert res.msgs is not None

    def test_broadcast_txcodespacerror(self, wallet, fee, mnemonics):
        """Tests that a that it captures the correct txcodespace error."""
        fail_swap = MsgSwap(
            trader=wallet.address, offer_coin=Coin("uluna", 1), ask_denom="bebo"
        )
        fail_vote = MsgExchangeRateVote(
            exchange_rate="603.899000000000000000",
            salt="0dff",
            denom="umnt",
            feeder=wallet.address,
            validator="terravaloper1vqnhgc6d0jyggtytzqrnsc40r4zez6tx99382w",
        )
        send = MsgSend(
            from_address=wallet.address,
            to_address=mnemonics[1]["address"],
            amount=Coins(uluna=1),
        )

        tx1 = wallet.create_and_sign_tx(send, send, fail_swap, send, fail_vote, fee=fee)
        with pytest.raises(TxCodespaceError) as excinfo:
            wallet.broadcast(tx1)
        err = excinfo.value
        assert err.codespace == "market"

        tx2 = wallet.create_and_sign_tx(
            send, fail_vote, send, send, fail_swap, fail_swap, fee=fee
        )
        with pytest.raises(TxCodespaceError) as excinfo:
            wallet.broadcast(tx2)
        err = excinfo.value
        assert err.codespace == "oracle"

    def test_make_proposal(self, wallet, fee):
        e = MsgSubmitProposal(
            content=ParameterChangeProposal(
                "testing params",
                "yay!",
                changes={
                    "distribution": {
                        "community_tax": Dec(0),
                        "base_proposer_reward": Dec(32),
                        "bonus_proposer_reward": Dec(22),
                        "withdraw_addr_enabled": True,
                    },
                    "staking": {
                        "unbonding_time": 33,
                        "max_validators": 9999,
                        "max_entries": 2323,
                        "bond_denom": "uluna",
                    },
                    "slashing": {
                        "max_evidence_age": 234234,
                        "signed_blocks_window": 1,
                        "min_signed_per_window": Dec(1),
                        "downtime_jail_duration": 1,
                        "SlashFractionDoubleSign": Dec(100),
                        "slash_fraction_downtime": Dec(213.123),
                    },
                    "treasury": {
                        "tax_policy": PolicyConstraints(
                            rate_min=Dec(0),
                            rate_max=Dec(100),
                            cap=Coin("unused", 0),
                            change_max=Dec(3),
                        ),
                        "reward_policy": PolicyConstraints(
                            rate_min=Dec(0),
                            rate_max=Dec(1023423340),
                            cap=Coin("unused", 0),
                            change_max=Dec(3),
                        ),
                        "seigniorage_burden_target": Dec("2342.234234"),
                        "mining_increment": Dec(23423423423.234234234234982),
                        "window_short": 50,
                        "window_long": 2,
                        "window_probation": 30,
                    },
                    "oracle": {
                        "vote_period": 345345,
                        "vote_threshold": Dec("2342.234333"),
                        "reward_band": Dec("234343"),
                        "reward_distribution_window": 345345,
                        "whitelist": ["abc", "bdc", "ttt"],
                        "slash_fraction": Dec(23423.232343),
                        "slash_window": 343311,
                        "min_valid_per_window": Dec(2342.234234),
                    },
                    "market": {
                        "pool_recovery_period": 234234234,
                        "base_pool": Dec(232323232),
                        "min_spread": Dec(343434),
                        "illiquid_tobin_tax_list": [{"denom": "item", "msg": "sdfsdf"}],
                    },
                    "gov": {
                        "deposit_params": {
                            "min_deposit": Coins(uluna=2, ukrw=5),
                            "max_deposit_period": 30434,
                        },
                        "voting_params": {"voting_period": 434243234},
                        "tallyparams": {
                            "quorum": Dec(234234.2334),
                            "threshold": Dec(23423.2323),
                            "veto": Dec(1232.234),
                        },
                    },
                },
            ),
            initial_deposit=Coins(uluna=10000000),
            proposer=wallet.address,
        )
        print(wallet.address)
        print(wallet.account_number, wallet.sequence)
        tx = wallet.create_and_sign_tx(e, fee=fee)
        res = wallet.broadcast(tx)




import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.msg.market import MsgSwap
from jigu.core.sdk import Coin
from jigu.error import InvalidAccAddress
from jigu.util.validation import is_acc_address
from testtools import assert_serdes_consistent, assert_serdes_exact, load_msg_examples


@pytest.fixture(scope="module")
def msg_examples(tdd):
    """Stores the __schema__ validation set for market messages."""
    return load_msg_examples(tdd, msg_types=[MsgSwap],)


class TestMsgSwap:
    def test_schema_valid(self, msg_examples):
        for m in msg_examples[MsgSwap]:
            assert_serdes_consistent(MsgSwap, m)
            assert_serdes_exact(MsgSwap, m)

    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(m=from_schema(MsgSwap.__schema__))
    def test_serdes_consistent(self, m):
        assert_serdes_consistent(MsgSwap, m)

    def test_matches_meta(self):
        assert MsgSwap.type == "market/MsgSwap"
        assert MsgSwap.action == "swap"

    @given(other=st.text())
    def test_constructor_validates_addresses(self, acc_address, other):
        """MsgSwap should validate `trader` to be AccAddress.
        """
        assume(not is_acc_address(other))
        with pytest.raises(InvalidAccAddress):
            MsgSwap(trader=other, offer_coin=Coin("uluna", 100), ask_denom="umnt")
        A = MsgSwap(trader=acc_address, offer_coin=Coin("uluna", 100), ask_denom="umnt")
        assert A.trader == acc_address



from __future__ import annotations

from dataclasses import dataclass

from jigu.core import AccAddress, Coin
from jigu.core.msg import StdMsg
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address

__all__ = ["MsgSwap"]


@dataclass
class MsgSwap(StdMsg):

    type = "market/MsgSwap"
    action = "swap"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^market/MsgSwap\Z"),
        value=S.OBJECT(
            trader=S.ACC_ADDRESS, offer_coin=Coin.__schema__, ask_denom=S.STRING
        ),
    )

    trader: AccAddress
    offer_coin: Coin
    ask_denom: str

    def __post_init__(self):
        self.trader = validate_acc_address(self.trader)

    @classmethod
    def from_data(cls, data: dict) -> MsgSwap:
        data = data["value"]
        return cls(
            trader=data["trader"],
            offer_coin=Coin.from_data(data["offer_coin"]),
            ask_denom=data["ask_denom"],
        )



from typing import List, Optional, Set, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import (
    AccAddress,
    Coin,
    Coins,
    Dec,
    ExchangeRatePrevote,
    ExchangeRateVote,
    ValAddress
)
from jigu.error import DenomNotFound
from jigu.util.serdes import JiguBox
from jigu.util.validation import validate_val_address

__all__ = ["OracleApi"]


class OracleApi(BaseApi):

    # for some reason, votes and prevotes for a certain (val, denom) pair will give STILL give you a list
    def votes(
        self, validator: Optional[ValAddress] = None, denom: Optional[str] = None
    ) -> Union[ApiResponse, List[ExchangeRateVote]]:
        if validator is not None and denom is not None:
            validator = validate_val_address(validator)
            res = self._api_get(f"/oracle/denoms/{denom}/votes/{validator}")
        elif validator:
            validator = validate_val_address(validator)
            res = self._api_get(f"/oracle/voters/{validator}/votes")
        elif denom:
            res = self._api_get(f"/oracle/denoms/{denom}/votes")
        else:
            raise ValueError("arguments validator and denom cannot both be None")
        return project(res, [ExchangeRateVote.deserialize(vote) for vote in res])

    def prevotes(
        self, validator: Optional[ValAddress] = None, denom: Optional[str] = None
    ) -> Union[ApiResponse, List[ExchangeRatePrevote]]:
        if validator is not None and denom is not None:
            validator = validate_val_address(validator)
            res = self._api_get(f"/oracle/denoms/{denom}/prevotes/{validator}")
        elif validator:
            validator = validate_val_address(validator)
            res = self._api_get(f"/oracle/voters/{validator}/prevotes")
        elif denom:
            res = self._api_get(f"/oracle/denoms/{denom}/prevotes")
        else:
            raise ValueError("arguments validator and denom cannot both be None")
        return project(
            res, [ExchangeRatePrevote.deserialize(prevote) for prevote in res]
        )

    def exchange_rates(self) -> Union[ApiResponse, Coins]:
        """Gets all exchange rates."""
        res = self._api_get("/oracle/denoms/exchange_rates")
        return project(res, Coins.from_data(res))

    def exchange_rate(self, denom: str) -> Union[ApiResponse, Coin]:
        """Gets the exchange rate of LUNA against one denomination."""
        res = self.exchange_rates()
        if denom not in res:
            raise DenomNotFound(
                f"denom '{denom}' not found, available denoms: {res.denoms}"
            )
        return project(res, res[denom])

    def active_denoms(self) -> Union[ApiResponse, Set[str]]:
        res = self._api_get("/oracle/denoms/actives")
        return project(res, set(res))

    def feeder_address_for(
        self, validator: ValAddress
    ) -> Union[ApiResponse, AccAddress]:
        validator = validate_val_address(validator)
        return self._api_get(f"/oracle/voters/{validator}/feeder")

    def misses_for(self, validator: ValAddress) -> Union[ApiResponse, int]:
        validator = validate_val_address(validator)
        res = self._api_get(f"/oracle/voters/{validator}/miss")
        return project(res, int(res))

    def params(
        self, key: Optional[str] = None
    ) -> Union[ApiResponse, dict, int, Dec, List[str]]:
        res = self._api_get("/oracle/parameters")
        p = JiguBox(
            res,
            box_recast={
                "vote_period": int,
                "vote_threshold": Dec,
                "reward_band": Dec,
                "reward_distribution_window": int,
                "slash_fraction": Dec,
                "slash_window": int,
                "min_valid_per_window": Dec,
            },
        )
        return project(res, p[key] if key else p)



import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core.auth.transaction import StdTx
from jigu.core.sdk import Coin
from testtools import assert_serdes_consistent, assert_serdes_exact, load_stdtx_examples


@pytest.fixture(scope="module")
def examples(tdd):
    # need to create the data by looking inside txinfo
    return load_stdtx_examples(tdd)


class TestStdTxSerdes:
    def test_schema_valid(self, examples):
        for tx in examples:
            assert_serdes_consistent(StdTx, tx)
            # assert_serdes_exact(StdTx, tx)
            # Cannot be exact due to inability to preserve json order

    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(tx=from_schema(StdTx.__schema__))
    def test_serdes_consistent(self, tx):
        assert_serdes_consistent(StdTx, tx)



import warnings
from typing import Optional, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import Coin, Dec
from jigu.util.serdes import JiguBox

__all__ = ["MarketApi"]


class MarketApi(BaseApi):
    def swap_rate(self, offer_coin: Coin, ask_denom: str) -> Union[ApiResponse, Coin]:
        if type(offer_coin.amount) != int:
            warnings.warn(
                f"Coin's amount will be converted to integer: {int(offer_coin.amount)} {offer_coin.denom}",
                SyntaxWarning,
            )
        params = {
            "offer_coin": f"{int(offer_coin.amount)}{offer_coin.denom}",
            "ask_denom": ask_denom,
        }
        res = self._api_get(f"/market/swap", params=params)
        return project(res, Coin.deserialize(res))

    def terra_pool_delta(self) -> Union[ApiResponse, Dec]:
        res = self._api_get("/market/terra_pool_delta")
        return project(res, Dec.deserialize(res))

    def params(self, key: Optional[str] = None) -> Union[ApiResponse, Dec, dict]:
        res = self._api_get("/market/parameters")
        p = JiguBox(res)
        p["pool_recovery_period"] = int(p["pool_recovery_period"])
        p["base_pool"] = Dec.deserialize(p["base_pool"])
        p["min_spread"] = Dec.deserialize(p["min_spread"])
        p["tobin_tax"] = Dec.deserialize(p["tobin_tax"])
        ill = p["illiquid_tobin_tax_list"]
        p["illiquid_tobin_tax_list"] = JiguBox({})
        for item in ill:
            p["illiquid_tobin_tax_list"][item["denom"]] = Dec(item["tax_rate"])
        return project(res, p[key] if key else p)



from __future__ import annotations

from dataclasses import dataclass
from typing import Type

from jigu.core import AccAddress, Coins
from jigu.core.msg import StdMsg
from jigu.core.proposal import PROPOSAL_TYPES, Content
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_acc_address

__all__ = ["MsgSubmitProposal", "MsgDeposit", "MsgVote"]


@dataclass
class MsgSubmitProposal(StdMsg):

    type = "gov/MsgSubmitProposal"
    action = "submit_proposal"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^gov/MsgSubmitProposal\Z"),
        value=S.OBJECT(
            content=S.ANY(*(pt.__schema__ for pt in PROPOSAL_TYPES.values())),
            initial_deposit=Coins.__schema__,
            proposer=S.ACC_ADDRESS,
        ),
    )

    content: Type[Content]
    initial_deposit: Coins
    proposer: AccAddress

    def __post_init__(self):
        self.proposer = validate_acc_address(self.proposer)
        self.initial_deposit = Coins(self.initial_deposit)

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitProposal:
        data = data["value"]
        p_type = PROPOSAL_TYPES[data["content"]["type"]]
        content = p_type.from_data(data["content"])
        return cls(
            content=content,
            initial_deposit=Coins.from_data(data["initial_deposit"]),
            proposer=data["proposer"],
        )


@dataclass
class MsgDeposit(StdMsg):

    type = "gov/MsgDeposit"
    action = "deposit"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^gov/MsgDeposit\Z"),
        value=S.OBJECT(
            proposal_id=S.STRING_INTEGER,
            depositor=S.ACC_ADDRESS,
            amount=Coins.__schema__,
        ),
    )

    proposal_id: int
    depositor: AccAddress
    amount: Coins

    def __post_init__(self):
        self.depositor = validate_acc_address(self.depositor)
        self.amount = Coins(self.amount)

    def msg_value(self) -> dict:
        d = dict(self.__dict__)
        d["proposal_id"] = str(self.proposal_id)
        return d

    @classmethod
    def from_data(cls, data: dict) -> MsgDeposit:
        data = data["value"]
        return cls(
            proposal_id=int(data["proposal_id"]),
            depositor=data["depositor"],
            amount=Coins.from_data(data["amount"]),
        )


@dataclass
class MsgVote(StdMsg):

    type = "gov/MsgVote"
    action = "vote"

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^gov/MsgVote\Z"),
        value=S.OBJECT(
            proposal_id=S.STRING_INTEGER,
            voter=S.ACC_ADDRESS,
            option=S.ONE(S.STRING, S.INTEGER),  # signing is different
        ),
    )

    EMPTY = "Empty"
    YES = "Yes"
    ABSTAIN = "Abstain"
    NO = "No"
    NO_WITH_VETO = "NoWithVeto"

    proposal_id: int
    voter: AccAddress
    option: str

    def __post_init__(self):
        self.voter = validate_acc_address(self.voter)

    def msg_value(self) -> dict:
        d = dict(self.__dict__)
        d["proposal_id"] = str(self.proposal_id)
        return d

    @classmethod
    def from_data(cls, data: dict) -> MsgVote:
        data = data["value"]
        return cls(
            proposal_id=int(data["proposal_id"]),
            voter=data["voter"],
            option=data["option"],
        )



from typing import List, Optional, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import (
    AccAddress,
    Coin,
    Delegation,
    Redelegation,
    UnbondingDelegation,
    ValAddress,
    Validator,
)
from jigu.core.denoms import uLuna
from jigu.util.serdes import JiguBox
from jigu.util.validation import validate_acc_address, validate_val_address

__all__ = ["StakingApi"]


class StakingApi(BaseApi):
    def delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
    ) -> Union[ApiResponse, List[Delegation]]:
        """Queries the delegation between a delegator and a validator."""
        if delegator is not None and validator is not None:
            delegator = validate_acc_address(delegator)
            validator = validate_val_address(validator)
            res = self._api_get(
                f"/staking/delegators/{delegator}/delegations/{validator}"
            )
            return project(res, [Delegation.deserialize(res)])
        elif delegator:
            delegator = validate_acc_address(delegator)
            res = self._api_get(f"/staking/delegators/{delegator}/delegations")
        elif validator:
            validator = validate_val_address(validator)
            res = self._api_get(f"/staking/validators/{validator}/delegations")
        else:
            raise TypeError("arguments delegator and validator cannot both be None")
        return project(res, [Delegation.deserialize(delgn) for delgn in res])

    def delegation(
        self, delegator: AccAddress, validator: ValAddress
    ) -> Union[ApiResponse, Delegation]:
        return self.delegations(delegator, validator)[0]

    def unbonding_delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
    ) -> Union[ApiResponse, List[UnbondingDelegation]]:
        if delegator is not None and validator is not None:
            delegator = validate_acc_address(delegator)
            validator = validate_val_address(validator)
            res = self._api_get(
                f"/staking/delegators/{delegator}/unbonding_delegations/{validator}"
            )
            return project(res, [UnbondingDelegation.deserialize(res)])
        elif delegator:
            delegator = validate_acc_address(delegator)
            res = self._api_get(
                f"/staking/delegators/{delegator}/unbonding_delegations"
            )
        elif validator:
            validator = validate_val_address(validator)
            res = self._api_get(
                f"/staking/validators/{validator}/unbonding_delegations"
            )
        else:
            raise TypeError("arguments delegator and validator cannot both be None")
        return project(res, [UnbondingDelegation.deserialize(delgn) for delgn in res])

    def unbonding_delegation(
        self, delegator: AccAddress, validator: ValAddress
    ) -> Union[ApiResponse, UnbondingDelegation]:
        return self.unbonding_delegations(delegator, validator)[0]

    def redelegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator_src: Optional[ValAddress] = None,
        validator_dst: Optional[ValAddress] = None,
    ) -> Union[ApiResponse, List[Redelegation]]:
        params = {}
        if delegator:
            delegator = validate_acc_address(delegator)
            params["delegator"] = delegator
        if validator_src:
            validator_src = validate_val_address(validator_src)
            params["validator_from"] = validator_src
        if validator_dst:
            validator_dst = validate_val_address(validator_dst)
            params["validator_to"] = validator_dst
        res = self._api_get(f"/staking/redelegations", params=params)
        return project(res, [Redelegation.deserialize(rd) for rd in res])

    def bonded_validators_for(
        self, delegator: AccAddress
    ) -> Union[ApiResponse, List[Validator]]:
        delegator = validate_acc_address(delegator)
        vs = self._api_get(f"/staking/delegators/{delegator}/validators")
        return [Validator.deserialize(v) for v in vs]

    def staking_txs_for(self, delegator: AccAddress) -> ApiResponse:
        delegator = validate_acc_address(delegator)
        return self._api_get(f"/staking/delegators/{delegator}/txs", unwrap=False)

    def validators(
        self, status: Optional[str] = None
    ) -> Union[ApiResponse, List[Validator]]:
        params = dict()
        if status is not None:
            params["status"] = status
        res = self._api_get("/staking/validators", params=params)
        return project(res, [Validator.deserialize(v) for v in res])

    def val_info_for(self, validator: ValAddress) -> Union[ApiResponse, Validator]:
        validator = validate_val_address(validator)
        res = self._api_get(f"/staking/validators/{validator}")
        return project(res, Validator.deserialize(res))

    def pool(
        self, key: Optional[str] = None
    ) -> Union[ApiResponse, Coin, JiguBox[str, Coin]]:
        res = self._api_get("/staking/pool")
        pool = JiguBox(
            {
                "bonded": Coin(uLuna, res["bonded_tokens"]),
                "not_bonded": Coin(uLuna, res["not_bonded_tokens"]),
            }
        )
        return project(res, pool[key] if key else pool)

    def params(self, key: Optional[str] = None) -> Union[ApiResponse, JiguBox]:
        res = self._api_get("/staking/parameters")
        p = JiguBox(res, box_recast={"unbonding_time": int})
        return project(res, p[key] if key else p)



import pytest

from jigu import Terra
from jigu.key.mnemonic import MnemonicKey
from jigu.core.sdk.coin import Coins
from jigu.core.msg.bank import MsgSend
from jigu.error import TxError
from jigu.core.auth.transaction import TxBroadcastResult


@pytest.fixture
def msg(wallet, mnemonics):
    recipient = mnemonics[2]["address"]
    return MsgSend(
        from_address=wallet.address, to_address=recipient, amount=Coins(uluna=1)
    )


class TestWallet:
    """Tests wallet transaction features."""

    def test_account_number(self, wallet):
        assert wallet.account_number == 189

    def test_sequence(self, wallet, msg, fee):
        """Sequence should increment after a transaction."""
        seq = wallet.sequence
        tx = wallet.create_and_sign_tx(msg, fee=fee)
        wallet.broadcast(tx)
        assert wallet.sequence == seq + 1

    def test_multiple_tx(self, wallet, msg):
        """Multiple TX's generated should have the same sequence number because the
        sequence number is live-fetched from the blockchain each time."""
        seq = wallet.sequence
        tx1 = wallet.create_tx(msg)
        tx2 = wallet.create_tx(msg)
        tx3 = wallet.create_tx(msg)

        assert tx1.sequence == seq
        assert tx2.sequence == seq
        assert tx3.sequence == seq

    def test_manual(self, wallet, msg):
        """Wallet can be used in manual-incrementing mode context, where the automatic
        fetching of sequence is paused for tx generation."""
        seq = wallet.sequence
        with wallet.manual() as w:
            tx1 = w.create_tx(msg)
            tx2 = w.create_tx(msg)
            _ = w.create_and_sign_tx(msg)  # make stdtx, and sequence is not known
            tx4 = w.create_tx(msg)
            assert w.sequence == seq  # wallet.sequence behavior should not change
            # this only applies to creating TX behavior.
        assert tx1.sequence == seq
        assert tx2.sequence == seq + 1
        assert tx4.sequence == seq + 3
        assert wallet.sequence == seq

        tx_after = wallet.create_tx(msg)
        tx_after_2 = wallet.create_tx(msg)

        # normal behavior should resume.
        assert tx_after.sequence == seq
        assert tx_after_2.sequence == seq

    def test_new_wallet_send(self, terra, wallet, fee):
        """A new wallet should be able to make transactions."""

        new_wallet = terra.wallet(MnemonicKey.generate())
        send = MsgSend(
            from_address=new_wallet.address,
            to_address=wallet.address,
            amount=Coins(uluna=1),
        )
        tx = new_wallet.create_and_sign_tx(send, fee=fee)
        with pytest.raises(TxError):
            new_wallet.broadcast(tx)

        tx = wallet.create_and_sign_tx(
            MsgSend(
                from_address=wallet.address,
                to_address=new_wallet.address,
                amount=Coins(uluna=10000000),
            ),
            fee=fee,
        )
        wallet.broadcast(tx)
        tx = new_wallet.create_and_sign_tx(send, fee=fee)
        res = new_wallet.broadcast(tx)
        assert isinstance(res, TxBroadcastResult)



from .account import AccountQuery
from .blocks import BlocksQuery
from .denom import DenomQuery
from .proposal import ProposalQuery
from .validator import ValidatorQuery
from .wallet import Wallet

__all__ = [
    "AccountQuery",
    "BlocksQuery",
    "DenomQuery",
    "ProposalQuery",
    "ValidatorQuery",
    "Wallet",
]



from __future__ import annotations

import re
from decimal import Decimal

from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

SDK_DEC_PREC = 18
SDK_DEC_UNIT = 10 ** SDK_DEC_PREC
SDK_DEC_REGEX_PATTERN = r"^(\-)?(\d+)(\.\d+)?\Z"
PATTERN = re.compile(SDK_DEC_REGEX_PATTERN)

__all__ = ["SDK_DEC_PREC", "Dec"]


class Dec(JsonSerializable, JsonDeserializable):

    __schema__ = S.STRING_WITH_PATTERN(SDK_DEC_REGEX_PATTERN)

    def __init__(self, arg):
        """BigInt-based Decimal representation with basic arithmetic operations with
        compatible Python numeric types (int, float, Decimal). Does not work with
        NaN, Infinity, +0, -0, etc. Serializes as a string with 18 points of decimal
        precision.
        """
        if isinstance(arg, float):
            arg = str("%f" % arg)
        if isinstance(arg, str):
            parts = PATTERN.match(arg)
            if parts is None:
                raise ValueError(f"Unable to parse Dec object from string: {arg}")
            self.i = int(parts.group(2)) * SDK_DEC_UNIT  # integer part
            if parts.group(3):
                fraction = (
                    parts.group(3).lstrip(".")[0:SDK_DEC_PREC].ljust(SDK_DEC_PREC, "0")
                )
                self.i += int(fraction)
            if parts.group(1) is not None:
                self.i *= -1
        elif isinstance(arg, int):
            self.i = arg * SDK_DEC_UNIT
        elif isinstance(arg, Dec):
            self.i = arg.i
        elif isinstance(arg, Decimal):
            whole = int(arg)
            fraction = arg % 1
            self.i = (whole * SDK_DEC_UNIT) + (fraction * SDK_DEC_UNIT)
        else:
            raise TypeError(f"Unable to create Dec object from given argument {arg}")
        # guarantee int
        self.i = int(self.i)

    @property
    def short_str(self):
        parity = "-" if self.i < 0 else ""
        frac = self.frac.rstrip("0")
        dot = "." if len(frac) > 0 else ""
        return f"{parity}{self.whole}{dot}{frac}"

    def __repr__(self):
        return f"Dec({self.short_str!r})"  # short representation

    def __str__(self) -> str:
        if self.i == 0:
            return "0." + SDK_DEC_PREC * "0"
        parity = "-" if self.i < 0 else ""
        return f"{parity}{self.whole}.{self.frac}"

    def __int__(self) -> int:
        int_part = abs(self.i) // SDK_DEC_UNIT
        int_part *= -1 if self.i < 0 else 1
        return int_part

    def __float__(self) -> float:
        # NOTE: This is not robust enough for: float(Dec(float)) to give the same output,
        # and should mainly be used as getting a rough value from the Dec object.
        return self.i / SDK_DEC_UNIT

    @property
    def parity(self) -> int:
        return -1 if self.i < 0 else 1

    @property
    def whole(self) -> str:
        return str(abs(self.i) // SDK_DEC_UNIT)

    @property
    def frac(self) -> str:
        return str(abs(self.i) % SDK_DEC_UNIT).rjust(SDK_DEC_PREC, "0")

    def to_data(self) -> str:
        return str(self)

    def _pretty_repr_(self, path: str = ""):
        return self.short_str

    def __eq__(self, other) -> bool:
        try:
            return self._i_binop(other, int.__eq__, "==")
        except TypeError:
            return False

    def __lt__(self, other) -> bool:
        if isinstance(other, Dec):
            return self.i < other.i
        return (Decimal(self.i) / SDK_DEC_UNIT) < other

    def __le__(self, other) -> bool:
        return self < other or self == other

    def __gt__(self, other) -> bool:
        if isinstance(other, Dec):
            return self.i > other.i
        return (Decimal(self.i) / SDK_DEC_UNIT) > other

    def __ge__(self, other) -> bool:
        return self > other or self == other

    def __add__(self, other) -> Dec:
        new_val = self._i_binop(other, int.__add__, "+")
        nd = Dec(0)
        nd.i = new_val
        return nd

    def __radd__(self, other):
        return Dec.__add__(Dec(other), self)

    def __sub__(self, other) -> Dec:
        new_val = self._i_binop(other, int.__sub__, "-")
        nd = Dec(0)
        nd.i = new_val
        return nd

    def __rsub__(self, other):
        return Dec.__sub__(Dec(other), self)

    def __mul__(self, other):
        if isinstance(other, Dec):
            new_val = self.i * other.i
        elif (
            isinstance(other, int)
            or isinstance(other, float)
            or isinstance(other, Decimal)
        ):
            new_val = self.i * Dec(other).i
        else:
            raise TypeError(
                f"unsupported operand types for *: 'Dec' and '{type(other)}'"
            )
        nd = Dec(0)
        nd.i = new_val // SDK_DEC_UNIT
        return nd

    def __rmul__(self, other):
        return Dec.__mul__(Dec(other), self)

    def __truediv__(self, other):
        try:
            # we need the decimal value's better precision.
            d1 = Decimal(self.i)
            d2 = Decimal(other.i)
            new_val = d1 / d2
        except AttributeError:
            new_val = self._i_binop(other, int.__truediv__, "/")
        nd = Dec(0)
        nd.i = int(new_val * SDK_DEC_UNIT)
        return nd

    def __rtruediv__(self, other):
        return Dec.__truediv__(Dec(other), self)

    def __floordiv__(self, other):
        return Dec(int(self / Dec(other)))

    def __neg__(self):
        x = Dec(self)
        x.i *= -1
        return x

    def __abs__(self):
        x = Dec(self)
        x.i = abs(x.i)
        return x

    def __pos__(self):
        return abs(self)

    @classmethod
    def from_data(cls, data: str) -> Dec:
        return cls(data)

    @classmethod
    def with_prec(cls, i: int, prec: int) -> Dec:
        """Replicates Dec.with_prec(i, prec)"""
        d = cls(0)
        i = int(i)
        d.i = i * 10 ** (SDK_DEC_PREC - int(prec))
        return d

    def _i_binop(self, other, binop, binop_name):
        """Helper method that tries to work with compatible number types by first converting
        them into Dec and working with their internal BigInt representation.
        """
        if isinstance(other, Dec):
            new_val = binop(self.i, other.i)
        elif (
            isinstance(other, int)
            or isinstance(other, float)
            or isinstance(other, Decimal)
        ):
            new_val = binop(self.i, Dec(other).i)
        else:
            raise TypeError(
                f"unsupported operand types for {binop_name}: 'Dec' and '{type(other)}'"
            )
        return new_val



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Union

from jigu.core import Coin, Dec
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = ["PolicyConstraints"]


@dataclass
class PolicyConstraints(JsonSerializable[dict], JsonDeserializable[dict]):

    __schema__ = S.OBJECT(
        rate_min=Dec.__schema__,
        rate_max=Dec.__schema__,
        cap=Coin.__schema__,
        change_max=Dec.__schema__,
    )

    rate_min: Dec
    rate_max: Dec
    cap: Coin
    change_max: Dec

    def clamp(self, prev_rate: Dec, new_rate: Dec):
        prev_rate = Dec(prev_rate)
        new_rate = Dec(new_rate)

        if new_rate < self.rate_min:
            new_rate = self.rate_min
        elif new_rate > self.rate_max:
            new_rate = self.rate_max

        delta = new_rate - prev_rate
        if new_rate > prev_rate:
            if delta > self.change_max:
                new_rate = prev_rate + self.change_max
        else:
            if abs(delta) > self.change_max:
                new_rate = prev_rate - self.change_max
        return new_rate

    @classmethod
    def from_data(cls, data: dict) -> PolicyConstraints:
        return cls(
            rate_min=Dec(data["rate_min"]),
            rate_max=Dec(data["rate_max"]),
            cap=Coin.from_data(data["cap"]),
            change_max=Dec(data["change_max"]),
        )



from typing import Dict, List, Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import AccAddress, Coin, Coins, Dec, Proposal
from jigu.core.denoms import uLuna
from jigu.util.serdes import JiguBox

__all__ = ["GovApi"]


class GovApi(BaseApi):
    """Interface for interacting with the Governance API."""

    def proposals(self) -> Union[ApiResponse, List[Proposal]]:
        """Get all proposals."""
        res = self._api_get("/gov/proposals")
        return project(res, [Proposal.deserialize(p) for p in res])

    def proposal(self, proposal_id: int) -> Union[ApiResponse, Proposal]:
        """Get a single proposal by ID."""
        res = self._api_get(f"/gov/proposals/{proposal_id}")
        return project(res, Proposal.deserialize(res))

    def proposer_for(self, proposal_id: int) -> Union[ApiResponse, AccAddress]:
        """Gets a proposal's proposer."""
        res = self._api_get(f"/gov/proposals/{proposal_id}/proposer")
        return project(res, res["proposer"])

    def deposits_for(self, proposal_id: int) -> Union[ApiResponse, Dict[str, Coins]]:
        """Get the proposal's deposits."""
        res = self._api_get(f"/gov/proposals/{proposal_id}/deposits")
        ds = res or []
        deposits = JiguBox({})
        for d in ds:
            deposits[d["depositor"]] = Coins.deserialize(d["amount"])
        return project(res, deposits)

    def votes_for(self, proposal_id: str) -> Union[ApiResponse, dict]:
        res = self._api_get(f"/gov/proposals/{proposal_id}/votes")
        vs = res or []
        votes = JiguBox({})
        for v in vs:
            votes[v["voter"]] = v["option"]
        return project(res, votes)

    def tally_for(self, proposal_id: str) -> Union[ApiResponse, Dict[str, Coin]]:
        res = self._api_get(f"/gov/proposals/{proposal_id}/tally")
        denoms = res or []
        tally = JiguBox({})
        for denom in denoms:
            tally[denom] = Coin(uLuna, int(res[denom]))
        return project(res, JiguBox(tally))

    def params(self, key: str = None):
        """Puts all the parameters together."""
        deposit = self.deposit_params()
        voting = self.voting_params()
        tally = self.tally_params()
        p = JiguBox(
            {"deposit_params": deposit, "voting_params": voting, "tally_params": tally}
        )

        return project(  # use response information of last entry, even if there is a delay
            tally, p[key] if key else p
        )

    def deposit_params(
        self, key: str = None
    ) -> Union[ApiResponse, Dict[str, Union[int, Coins]]]:
        res = self._api_get(f"/gov/parameters/deposit")
        p = JiguBox(res)
        p["min_deposit"] = Coins.deserialize(p["min_deposit"])
        p["max_deposit_period"] = int(p["max_deposit_period"])
        return project(res, p[key] if key else p)

    def voting_params(self, key: str = None) -> Union[ApiResponse, Dict[str, int]]:
        res = self._api_get(f"/gov/parameters/voting")
        p = JiguBox(res, box_recast={"voting_period": int})
        return project(res, p[key] if key else p)

    def tally_params(self, key: str = None) -> Union[ApiResponse, Dict[str, Dec]]:
        res = self._api_get(f"/gov/parameters/tallying")
        p = JiguBox(res, box_recast={"quorum": Dec, "threshold": Dec, "veto": Dec})
        return project(res, p[key] if key else p)



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Any, Dict, Type

from jigu.core import Coin, Coins, Timestamp
from jigu.core.denoms import uLuna
from jigu.util.serdes import JiguBox, JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S

__all__ = [
    "ProposalStatus",
    "Content",
    "PROPOSAL_TYPES",
    "Proposal",
    "TextProposal",
    "ParameterChangeProposal",
    "CommunityPoolSpendProposal",
    "TaxRateUpdateProposal",
    "RewardWeightUpdateProposal",
]


class ProposalStatus(str):
    NIL = ""
    DEPOSIT_PERIOD = "DepositPeriod"
    VOTING_PERIOD = "VotingPeriod"
    PASSED = "Passed"
    REJECTED = "Rejected"
    FAILED = "Failed"


class Content(JsonSerializable, JsonDeserializable, metaclass=abc.ABCMeta):

    __schema__ = S.OBJECT(
        type=S.STRING, value=S.OBJECT(title=S.STRING, description=S.STRING)
    )

    @property
    @abc.abstractmethod
    def type(self):
        raise NotImplementedError

    def proposal_value(self):
        return dict(self.__dict__)

    def to_data(self) -> Dict[str, Any]:
        return {"type": self.type, "value": self.proposal_value()}


from .distribution import CommunityPoolSpendProposal  # isort:skip
from .gov import TextProposal  # isort:skip
from .params import ParameterChangeProposal  # isort:skip
from .treasury import RewardWeightUpdateProposal, TaxRateUpdateProposal  # isort:skip


PROPOSAL_TYPES = {
    "gov/TextProposal": TextProposal,
    "params/ParameterChangeProposal": ParameterChangeProposal,
    "distribution/CommunityPoolSpendProposal": CommunityPoolSpendProposal,
    "treasury/TaxRateUpdateProposal": TaxRateUpdateProposal,
    "treasury/RewardWeightUpdateProposal": RewardWeightUpdateProposal,
}


@dataclass
class Proposal(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        content=Content.__schema__,
        id=S.STRING_INTEGER,
        proposal_status=S.STRING,
        final_tally_result=S.OPTIONAL(
            S.OBJECT(  # this gets marshalled into Coin
                yes=S.STRING_INTEGER,
                abstain=S.STRING_INTEGER,
                no=S.STRING_INTEGER,
                no_with_veto=S.STRING_INTEGER,
            )
        ),
        submit_time=Timestamp.__schema__,
        deposit_end_time=Timestamp.__schema__,
        total_deposit=Coins.__schema__,
        voting_start_time=Timestamp.__schema__,
        voting_end_time=Timestamp.__schema__,
    )

    content: Type[Content]
    id: int
    proposal_status: str
    final_tally_result: JiguBox[str, Coin]
    submit_time: Timestamp
    deposit_end_time: Timestamp
    total_deposit: Coins
    voting_start_time: Timestamp
    voting_end_time: Timestamp

    def to_data(self) -> dict:
        d = JiguBox(self.__dict__)
        d.id = str(d.id)
        for x in d.final_tally_result:
            d.final_tally_result[x] = d.final_tally_result[x].amount
        return d

    @property
    def pretty_data(self):
        d = dict(self.__dict__)
        proposal_id = d.pop("id")
        content = d.pop("content")
        ix = [
            ("id", proposal_id),
            ("type", content.type),
            *list(content.pretty_data),
            *list(d.items()),
        ]
        return ix

    @classmethod
    def from_data(cls, data: dict) -> Proposal:
        final_tally_result = data["final_tally_result"]
        for key in final_tally_result:
            final_tally_result[key] = Coin(uLuna, int(final_tally_result[key]))
        p_type = PROPOSAL_TYPES[data["content"]["type"]]
        content = p_type.from_data(data["content"])
        return cls(
            content=content,
            id=int(data["id"]),
            proposal_status=ProposalStatus(data["proposal_status"]),
            final_tally_result=JiguBox(final_tally_result),
            submit_time=Timestamp.from_data(data["submit_time"]),
            deposit_end_time=Timestamp.from_data(data["deposit_end_time"]),
            total_deposit=Coins.from_data(data["total_deposit"]),
            voting_start_time=Timestamp.from_data(data["voting_start_time"]),
            voting_end_time=Timestamp.from_data(data["voting_end_time"]),
        )



from datetime import datetime

import pytest
from hypothesis import assume, given
from hypothesis_jsonschema import from_schema

from jigu.core.sdk.timestamp import Timestamp
from testtools import assert_serdes_consistent, assert_serdes_exact


@pytest.mark.sdk
class TestTimestampSerdes:

    # TODO: write better regex so it is consistent.
    # @pytest.mark.serdes
    # @given(m=from_schema(Timestamp.__schema__))
    # def test_serdes_consistent(self, m):
    #     assert_serdes_consistent(Timestamp, m)

    def test_frac_second(self):
        examples = [
            "2019-04-24T06:05:20Z",
            "2019-04-24T06:05:10Z",
            "2019-06-02T22:23:38.000400591Z",
            "2020-02-24T03:48:05.950000000Z",
            "2020-02-07T12:41:42.636901000Z",
            "2020-02-22T01:05:40.237804060Z",
            "2020-02-04T13:26:16.364509600Z",
        ]
        for x in examples:
            ts = Timestamp.deserialize(x).to_data()
            assert ts == x

    def test_frac_second_ns_precision(self):
        examples = [
            (
                "2019-04-24T06:05:20.123456789123456789Z",
                "2019-04-24T06:05:20.123456789Z",
            ),
            ("2019-04-24T06:05:20.12Z", "2019-04-24T06:05:20.12Z",),
            ("2019-04-24T06:05:20.12345Z", "2019-04-24T06:05:20.12345Z",),
        ]
        for x in examples:
            ts = Timestamp.deserialize(x[0]).to_data()
            assert ts == x[1]

    def test_from_datetime(self):
        dt = datetime.now()
        ts = Timestamp.from_datetime(dt)
        dt_format = dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        ts_format = str(ts)
        assert dt_format == ts_format
        assert dt == ts

    def test_cmp(self):
        ts0 = Timestamp(2004, 12, 18, 19, 54, 22, 333333000, 9)
        ts1 = Timestamp(2004, 12, 18, 19, 54, 22, 333333333, 9)
        ts2 = Timestamp(2004, 12, 18, 19, 54, 22, 333333999, 9)
        ts3 = Timestamp(2004, 12, 18, 19, 54, 22, 533333999, 9)
        dt = datetime(2004, 12, 18, 19, 54, 22, 333333)
        dt2 = datetime(2004, 12, 18, 19, 54, 22, 433333)

        assert dt == ts0
        assert dt != ts1
        assert dt != ts2

        assert ts0 != ts1
        assert ts0 != ts2
        assert ts1 != ts2

        assert ts0 < ts1
        assert ts1 < ts2
        assert ts2 < dt2

        assert ts2 > dt
        assert ts1 > dt
        assert ts2 > ts1

        assert dt2 > ts2
        assert dt < ts3

        in_order = [dt, ts0, ts1, ts2, dt2, ts3]
        assert sorted([ts3, dt2, ts0, dt, ts1, ts2]) == in_order



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Union

from jigu.core import Coin, Dec, ValAddress
from jigu.util.serdes import JsonDeserializable, JsonSerializable
from jigu.util.validation import Schemas as S
from jigu.util.validation import validate_val_address

__all__ = ["ExchangeRatePrevote", "ExchangeRateVote"]


@dataclass
class ExchangeRateVote(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        denom=S.STRING, voter=S.VAL_ADDRESS, exchange_rate=Dec.__schema__
    )

    exchange_rate: Coin
    denom: str
    voter: ValAddress

    def __post_init__(self):
        self.voter = validate_val_address(self.voter)

    def to_data(self) -> Dict[str, Union[Dec, str, ValAddress]]:
        return {
            "exchange_rate": str(self.exchange_rate.amount),
            "denom": self.denom,
            "voter": self.voter,
        }

    @classmethod
    def from_data(cls, data) -> ExchangeRateVote:
        xr = Coin(data["denom"], data["exchange_rate"])
        return cls(exchange_rate=xr, denom=xr.denom, voter=data["voter"])


@dataclass
class ExchangeRatePrevote(JsonSerializable, JsonDeserializable):

    __schema__ = S.OBJECT(
        hash=S.STRING,
        denom=S.STRING,
        voter=S.VAL_ADDRESS,
        submit_block=S.STRING_INTEGER,
    )

    hash: str
    denom: str
    voter: ValAddress
    submit_block: int

    def __post_init__(self):
        self.voter = validate_val_address(self.voter)
        self.submit_block = int(self.submit_block)

    def to_data(self) -> Dict[str, str]:
        return {
            "hash": self.hash,
            "denom": self.denom,
            "voter": self.voter,
            "submit_block": str(self.submit_block),
        }

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> ExchangeRatePrevote:
        return cls(
            hash=data["hash"],
            denom=data["denom"],
            voter=data["voter"],
            submit_block=int(data["submit_block"]),
        )



from __future__ import annotations

import jigu.client
import jigu.client.object_query
from jigu.core import AccAddress, Proposal

__all__ = ["ProposalQuery"]


class ProposalQuery(object):
    def __init__(self, terra: jigu.client.terra.Terra, proposal_id: int):
        self.terra = terra
        self.proposal_id = proposal_id

    def __repr__(self):
        return f"ProposalQuery({self.proposal_id!r}) -> {self.terra}"

    def info(self) -> Proposal:
        return self.terra.gov.proposal(self.proposal_id)

    def proposer(self) -> AccAddress:
        return self.terra.gov.proposer_for(self.proposal_id)

    def deposits(self):
        return self.terra.gov.deposits_for(self.proposal_id)

    def votes(self):
        return self.terra.gov.votes_for(self.proposal_id)

    def tally(self):
        return self.terra.gov.tally_for(self.proposal_id)



"""Low-level transaction API"""
import json
from typing import Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import Coin, Coins, StdFee, StdSignMsg, StdTx, TxBroadcastResult, TxInfo
from jigu.error import RpcError, TxError, get_codespace_error
from jigu.util import hash_amino

# TODO: add tx transformer like blocks.transformer() to accomodate for custom messages
# ideal usage: terra.tx.transformer = my_transformer

__all__ = ["TxApi"]


class TxApi(BaseApi):
    """Low-level transactions API that interacts directly with the Tendermint logic of
    the node. Handles things that require on-chain connection such as hashing, amino-encoding,
    broadcasting, and transaction lookups.
    """

    def tx_info(self, txhash: Union[str]) -> Union[ApiResponse, TxInfo]:
        """Get the information for a transaction by its hash.

        :param txhash: transaction hash to lookup
        :type txhash: str

        Returns:
            Transaction info if found, otherwise `None`.
        """
        # TODO: add tx schema checking / error handling?
        res = self._api_get(f"/txs/{txhash}", unwrap=False)
        return project(res, TxInfo.deserialize(res))

    def _tx_info_threaded(self, txhashes):
        """Threaded version of tx_info, takes in a list of transaction hashes instead of
        just one transaction hash."""
        paths = [f"/txs/{txhash}" for txhash in txhashes]
        responses = self.lcd.get_threaded(paths)
        api_responses = [  # turn into API responses
            self._handle_response(resp, unwrap=False) for resp in responses
        ]
        return [project(res, TxInfo.deserialize(res)) for res in api_responses]

    def estimate_fee(
        self,
        tx: Union[StdSignMsg, StdTx],
        gas_prices: Coins = None,
        gas_adjustment: Union[float, str] = None,
    ) -> Union[ApiResponse, StdFee]:
        """Estimate a transaction's fee by simulating it in the node.

        Args:
            tx (Union[StdSignMsg, StdTx]): Transaction to calculate fee for.
            gas_prices (Coins, optional): Gas prices to use for estimation.
            gas_adjustment (Union[float, str], optional, default=1.0): Multiplicative
                factor to adjust for potential error.

        Returns:
            Estimated fee for the transaction.
        """
        if gas_prices is None:
            gas_prices = self.terra.gas_prices
        if gas_adjustment is None:
            gas_adjustment = self.terra.gas_adjustment
        if isinstance(tx, StdSignMsg):
            tx = tx.to_tx()
        tx.fee = StdFee(
            gas=0,  # simulation mode
            amount=Coins(
                Coin(d, 1) for d in gas_prices.denoms
            ),  # set 1 per denom we are including
        )
        data = {
            "tx": tx,
            "gas_adjustment": str(gas_adjustment),
            "gas_prices": gas_prices,
        }
        res = self._api_post("/txs/estimate_fee", data=data)
        gas = int(res["gas"])
        fees = Coins.deserialize(res["fees"])
        return project(res, StdFee(gas=gas, amount=fees))

    def encode(self, tx: StdTx) -> Union[ApiResponse, str]:
        """Generate the transaction's Amino enconding.

        Args:
            tx (StdTx): The transaction to encode.

        Returns:
            Base64-encoded string.
        """
        res = self._api_post("/txs/encode", data=tx, unwrap=False)
        return project(res, res["tx"])

    def hash(self, tx: StdTx) -> Union[ApiResponse, str]:
        """Gets the hash of a StdTx. This is in the Tx API due to the fact that Jigu
        must contact the server first, as Amino is not yet available on Python."""
        res = self.encode(tx)
        return project(res, hash_amino(res))

    def broadcast(
        self, tx: StdTx, mode: str = "block"
    ) -> Union[ApiResponse, TxBroadcastResult]:
        """Broadcast a signed transaction.

        :param tx: The signed transaction to broadcast.
        :param mode: One of "block", "sync", "async"
        """
        if mode not in ["block", "sync", "async"]:
            raise ValueError(
                f"mode '{mode}' is not legal; mode can only be 'block', 'sync', or 'async'."
            )
        data = {"tx": tx.to_data()["value"], "mode": mode}
        try:
            res = self._api_post("/txs", data=data, unwrap=False)
        except RpcError as e:
            raise TxError(e.message, tx)
        if "code" in res:  # status code 200, but TxError
            err = json.loads(res["raw_log"])
            # There are 2 types of Codespace errors that will be caught here:
            # 1) Message error : originating from a message
            # 2) Transaction : stuff like out of gas, insufficient fee, etc.
            #
            # If message error, `err` will be a list with the following structure:
            #
            # {'msg_index': 0, 'success': False, 'log': '{"codespace":"market","code":2,
            # "message":"No price registered with the oracle for asset: usdr"}', 'events':
            # [{'type': 'message', 'attributes': [{'key': 'action', 'value': 'swap'}]}]}
            #
            # Otherwise, `err` will be like (example).
            #
            # {"codespace":"market","code":2,
            # "message":"No price registered with the oracle for asset: usdr"}
            if isinstance(
                err, list
            ):  # encountered a msg-error; find the offending msg (there will only be 1)
                err = json.loads(err[-1]["log"])  # last one was the one responsible
            raise get_codespace_error(
                err["codespace"],
                err["code"],
                err["message"],
                was_from_tx=True,
                tx=tx,
                broadcast_result=TxBroadcastResult.from_data(res, tx),
            )
        retval = TxBroadcastResult.from_data(res, tx)
        return project(res, retval)



"""Classes for handling logic of block fetching."""

from multiprocessing.pool import ThreadPool
from typing import Union

from jigu.client.lcd.api import project
from jigu.core import Block
from jigu.transform.block import FetchTxInfoBlockTransformer

# TODO: rename to BlocksAPI ?

__all__ = ["BlocksQuery", "BlocksRange"]


class BlocksRange(object):
    def __init__(
        self, blocks, start_height: int = None, stop_height: int = None, step: int = 1
    ):
        """Creates a generator that iterates over [start_height, until_height]. Note that
        this goes against the convention that ranges should be [start, end); however humans
        think in inclusive intervals for block height.
        """
        self.blocks = blocks
        self.start_height = start_height or 1
        self.stop_height = stop_height or len(self.blocks)
        self.step = step
        self._range = range(self.start_height, self.stop_height + 1, self.step)

    def __repr__(self):
        s = self
        return f"BlocksRange(start_height={s.start_height!r}, stop_height={s.stop_height!r}, step={s.step!r})"

    def __iter__(self):
        yield from self.__next__()

    def __next__(self):
        for i in self._range:
            yield self.blocks.at(i)

    def threads(self, num_threads: int = 1):
        """Creates a multithreaded iterator to fetch blocks more quickly."""
        return ThreadPool(num_threads).imap(lambda i: self.blocks.at(i), self._range)


class BlocksQuery(object):
    """Manages how blocks are retrieved and parsed from the blockchain."""

    def __init__(self, terra):
        self.terra = terra
        self.transform_after_fetch = True
        self.transformer = FetchTxInfoBlockTransformer(terra)

    def _normalize_index(self, index: int):
        """Changes negative indices to positive ones."""
        if index < 0:
            return len(self) + index
        else:
            return index

    def __getitem__(self, key: Union[int, slice]):
        if isinstance(key, int):
            return self.at(self._normalize_index(key) + 1)
        elif isinstance(key, slice):
            start = self._normalize_index(key.start) + 1 if key.start else None
            stop = self._normalize_index(key.stop) + 1 if key.stop else len(self)
            step = key.step if key.step else 1
            return self.range(start, stop, step)
        else:
            raise TypeError(
                "invalid type for blocks[key], `key` must be <int> or <slice>"
            )

    def at(self, height: int):
        """Get the raw block data from the Tendermint API and deserializes it into a
        Block object. Applies the transformer(block) function to the block
        before returning.
        """
        res = self.terra._tendermint.block(height)  # block data
        block = Block.deserialize(res)
        if self.transform_after_fetch:
            block = self.transformer(block)
        return project(res, block)

    def range(
        self,
        start_height: int = None,
        stop_height: int = None,
        step: int = 1,
        threads: int = 1,
    ):
        # this mimics the python behavior: range(4) = range(0, 4); range(4, 8) = range(4, 8)
        result = BlocksRange(self, start_height, stop_height, step)
        if start_height is None:
            result = BlocksRange(self, None, None, step)
        if stop_height is None:
            result = BlocksRange(self, 1, start_height, step)
        if threads < 2:
            return result
        else:
            return result.threads(threads)

    def __call__(
        self, start_height: int, stop_height: int, step: int = 1, threads: int = 1
    ):
        return self.range(start_height, stop_height, step, threads)

    def __len__(self):
        res = self.terra._tendermint.block()  # get latest block
        return project(res, int(res.block.header.height))

    def threads(self, num_threads: int = 1):
        return self.range().threads(num_threads)



"""Quick and dirty scripts to load testing data and check serdes consistency and exactness."""

import json
import os
from pathlib import Path

import pytest

from jigu.client.lcd.lcdrequest import LcdRequest


def load_data(testdatadir: str, filepath: str):
    return json.load(open(os.path.join(testdatadir, filepath), "r"))


def extract_msgs(txinfos: list, whitelist: list = None):
    msgs = []
    for txinfo in txinfos:
        msgs += txinfo["tx"]["value"]["msg"]
    if whitelist:
        msgs = filter(lambda m: m["type"] in whitelist, msgs)
    return msgs


def load_msg_examples(testdatadir: str, msg_types: list) -> dict:
    examples = dict()
    for mt in msg_types:
        data = load_data(testdatadir, f"txinfo_by_msg/{mt.type}.json")
        examples[mt] = extract_msgs(data["txs"], whitelist=[mt.type])
    return examples


def load_stdtx_examples(testdatadir: str) -> dict:
    examples = []
    files = Path(os.path.join(testdatadir, "txinfo_by_msg")).rglob("*.json")
    for f in files:
        data = json.load(open(f, "r"))
        txinfos = data["txs"]
        for txinfo in txinfos:
            examples.append(txinfo["tx"])
    return examples


# API testing tools


class LcdRequestTest(Exception):
    """This helps us break out and report the request so we can
    check its value before making the actual request. To be used
    with the proper middleware installation in LcdClient.
    """

    def __init__(self, request: LcdRequest):
        Exception.__init__(self)
        self.method = request.method
        self.url = request.url
        self.kwargs = request.kwargs


def lcd_request_test_middleware(client, request: LcdRequest):
    """Exits execution before an actual request is made against the server.
    Used to check a generated request's value for testing."""
    raise LcdRequestTest(request)


## Custom assertions


def assert_serdes_consistent(klass, item):
    """Will attempt to deserialize an item, and check if serializing and
    deserializing again produces an equal object.
    """
    x = klass.deserialize(item)
    y = klass.deserialize(json.loads(x.to_json()))
    assert x == y


def assert_serdes_exact(klass, item):
    """Will attempt to deserialize an item, and check if serializing to JSON produces
    equivalent JSON.
    """
    x = klass.deserialize(item)
    output_json = json.loads(x.to_json())
    assert item == output_json



from collections import defaultdict
from typing import Dict, Iterable, List

from toolz import valmap

from jigu.core.event import Event
from jigu.util.pretty import PrettyPrintable
from jigu.util.serdes import JiguBox

__all__ = ["EventsQuery"]


class EventsQuery(PrettyPrintable):
    """Convenience query structure around a list of events.

    events[n] (n = int) --> n-th event
    events["event-type"] --> list of events of that type
    events.event_type --> list of events of that type
    for e in events: --> iterate over events
    "send" in events: --> whether events of this type are in collection
    """

    def __init__(self, events: Iterable[Event]):
        self.events = list(events)
        by_type = defaultdict(list)  # TODO: remove double representation..
        for e in self.events:
            by_type[e.type].append(e)
        self.by_type: JiguBox[str, List[Event]] = JiguBox(by_type)

    def __repr__(self) -> str:
        r = ", ".join(f"{t}: {n}" for t, n in self.types.items())
        if r == "":
            r = "(empty)"
        return f"<EventsQuery {r}>"

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.events[key]
        elif isinstance(key, slice):
            return EventsQuery(self.events[key])
        elif isinstance(key, str):
            # TODO: refactor by the example of MsgInfosQuery
            if key in self.by_type:
                return EventsQuery(self.by_type[key])
            else:
                return EventsQuery([])
        else:
            raise KeyError(
                f"unusable key {key} for EventsQuery, use 'int', 'slice', or 'str'"
            )

    def __getattr__(self, name):
        return self[name]

    def __len__(self) -> int:
        return sum(map(len, self.by_type.values()))

    def __iter__(self):
        return iter(self.events)

    @property
    def types(self) -> Dict[str, int]:
        return valmap(len, self.by_type)

    def __contains__(self, item):
        return item in self.by_type

    def to_dict(self) -> dict:  # for printing
        return self.by_type.to_dict()

    @property
    def pretty_data(self):
        return self.types.items()
